from math import sin

import requests
import numpy as np

from ..proto.encode import (
    encode_power_measurement,
    encode_teros12_measurement,
    encode_teros21_measurement,
    encode_bme280_measurement,
)

from ..proto.sensor import encode_repeated_sensor_measurements


class NodeSimulator:
    """Simulation class to simulate measurements for different sensors"""

    # temporary storage for measurements to be uploaded
    measurement_buffer: list[bytes] = []
    # all measurements uploaded
    measurements: list[bytes] = []
    # all responses
    responses: list[str] = []

    # metrics for uploads
    metrics: dict[str, int] = {
        "total_requests": 0,
        "failed_requests": 0,
        "successful_requests": 0,
    }

    latency: list[float] = []

    def __init__(self, cell: int, logger: int, sensors: list[str], fn=sin):
        self.cell = cell
        self.logger = logger
        self.sensors = sensors
        self.fn = fn

    def __str__(self):
        """String representation of the simulation class

        Shows the current upload metrics
        """
        avg = np.array(self.latency).mean()

        last = 0
        if len(self.latency) > 0:
            last = self.latency[-1]

        return "total: {}, failed: {}, avg (ms): {}, last (ms): {}".format(
            self.metrics["total_requests"],
            self.metrics["failed_requests"],
            avg * 100,
            last * 100,
        )

    def send_next(self, url: str) -> bool:
        """Sends measurements to a dirtviz instance

        Args:
            url: URL of the dirtviz instance

        Returns:
            True if there are measurements to send, False otherwise
        """

        # get next measurement
        try:
            meas = self.measurements.pop()
        except IndexError as _:
            return False

        headers = {"Content-Type": "application/octet-stream"}
        result = requests.post(url, data=meas, headers=headers)

        # store result
        self.responses.append(result.text)
        self.metrics["total_requests"] += 1
        if result.status_code == 200:
            self.metrics["successful_requests"] += 1
        else:
            self.metrics["failed_requests"] += 1
        self.latency.append(result.elapsed.total_seconds())

        return True

    def measure(self, ts: int):
        """Simulate measurements

        Args:
            ts: Timestamp of the measurement

        Returns:
            None
        """

        if "power" in self.sensors:

            voltage = self.fn(ts) * 2
            current = self.fn(ts) * 0.5

            meas = encode_power_measurement(
                ts=ts,
                cell_id=self.cell,
                logger_id=self.logger,
                voltage=voltage,
                current=current,
            )
            self.measurements.append(meas)
            self.measurement_buffer.append(meas)

        if "teros12" in self.sensors:

            vwc_raw = self.fn(ts) * 300 + 2500
            vwc_adj = self.fn(ts) * 0.05 + 0.2
            temp = self.fn(ts) * 5 + 25
            ec = self.fn(ts) * 4 + 15

            meas = encode_teros12_measurement(
                ts=ts,
                cell_id=self.cell,
                logger_id=self.logger,
                vwc_raw=vwc_raw,
                vwc_adj=vwc_adj,
                temp=temp,
                ec=int(ec),
            )
            self.measurements.append(meas)
            self.measurement_buffer.append(meas)

        if "teros21" in self.sensors:

            matric_pot = self.fn(ts) * 200 + 1000
            temp = self.fn(ts) * 5 + 25

            meas = encode_teros21_measurement(
                ts=ts,
                cell_id=self.cell,
                logger_id=self.logger,
                matric_pot=matric_pot,
                temp=temp,
            )
            self.measurements.append(meas)
            self.measurement_buffer.append(meas)

        if "bme280" in self.sensors:

            temp = self.fn(ts) * 50 + 250
            humidity = self.fn(ts) * 200 + 2000
            pressure = self.fn(ts) * 2000 + 43000

            meas = encode_bme280_measurement(
                ts=ts,
                cell_id=self.cell,
                logger_id=self.logger,
                temperature=int(temp),
                humidity=int(humidity),
                pressure=int(pressure),
            )
            self.measurements.append(meas)
            self.measurement_buffer.append(meas)


class NodeSimulatorGeneric:
    """Simulation class to simulate measurements for different sensors"""

    # temporary storage for measurements to be uploaded
    measurement_buffer: list[bytes] = []
    # all measurements uploaded
    measurements: list[bytes] = []
    # all responses
    responses: list[str] = []

    # metrics for uploads
    metrics: dict[str, int] = {
        "total_requests": 0,
        "failed_requests": 0,
        "successful_requests": 0,
    }

    latency: list[float] = []

    def __init__(
        self, cell: int, logger: int, sensors: list[str], _min=-1, _max=1, fn=sin
    ):
        """Initializes the simulation class.

        Args:
            cell: Cell ID of the node.
            logger: Logger ID of the node.
            sensors: List of sensors to simulate.
            _min: Minimum value for the simulated sensor data.
            _max: Maximum value for the simulated sensor data.
            fn: Function to generate the simulated sensor data.
        """

        self.cell = cell
        self.logger = logger
        self.sensors = sensors
        self.fn = fn
        self._min = _min
        self._max = _max

    def __str__(self):
        """String representation of the simulation class

        Shows the current upload metrics
        """
        avg = np.array(self.latency).mean()

        last = 0
        if len(self.latency) > 0:
            last = self.latency[-1]

        return "total: {}, failed: {}, avg (ms): {}, last (ms): {}".format(
            self.metrics["total_requests"],
            self.metrics["failed_requests"],
            avg * 100,
            last * 100,
        )

    def send_next(self, url: str) -> bool:
        """Sends measurements to a dirtviz instance

        Args:
            url: URL of the dirtviz instance

        Returns:
            True if there are measurements to send, False otherwise
        """

        # get next measurement
        try:
            meas = self.measurements.pop()
        except IndexError as _:
            return False

        headers = {"Content-Type": "application/octet-stream"}
        result = requests.post(url, data=meas, headers=headers)

        # store result
        self.responses.append(result.text)
        self.metrics["total_requests"] += 1
        if result.status_code == 200:
            self.metrics["successful_requests"] += 1
        else:
            self.metrics["failed_requests"] += 1
        self.latency.append(result.elapsed.total_seconds())

        return True

    def measure(self, ts: int):
        """Simulate measurements

        Args:
            ts: Timestamp of the measurement
        """

        meas = {
            "meta": {
                "ts": ts,
                "loggerId": self.logger,
                "cellId": self.cell,
            },
            "measurements": [],
        }

        scale = (self._max - self._min) / 2
        offset = (self._max + self._min) / 2

        for s in self.sensors:
            meas["measurements"].append(
                {
                    "type": s,
                    "decimal": self.fn(ts) * scale + offset,
                }
            )

        serialized = encode_repeated_sensor_measurements(meas)
        self.measurement_buffer.append(serialized)



from .encode import (
    encode_response,
    encode_power_measurement,
    encode_teros12_measurement,
    encode_phytos31_measurement,
    encode_teros21_measurement,
    encode_user_configuration,
)

from .decode import (
    decode_response,
    decode_measurement,
    decode_user_configuration,
)

from .esp32 import (
    encode_esp32command,
    decode_esp32command,
)

__all__ = [
    "encode_response",
    "encode_power_measurement",
    "encode_teros12_measurement",
    "encode_phytos31_measurement",
    "encode_teros21_measurement",
    "decode_response",
    "decode_measurement",
    "encode_user_configuration",
    "decode_user_configuration",
    "encode_esp32command",
    "decode_esp32command",
]



"""
@brief PyQt5 GUI Application for Configuring User Settings

This module provides a PyQt5-based graphical interface for configuring user settings.
The application allows users to input configuration details, including Logger ID, Cell ID,
Upload Method (WiFi or LoRa), Upload Interval, Enabled Sensors, and Calibration parameters
for voltage and current (V/I Slope and Offset).

Key features:
- **Save and Load**: Users can save configurations to a file or load previous configurations for easy reuse.
- **Real-time Configuration**: By pressing the "Send Configuration" button, the settings are serialized with Protobuf
  and transmitted over UART to the STM32 for direct application.

@file user_config.py
@author Ahmed Hassan Falah
@date 2024-10-10
"""

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import QInputDialog
import json
import os
import sys
import serial
import serial.tools.list_ports
import re  # For validating URL input
from ..proto import encode_user_configuration, decode_user_configuration


class Ui_MainWindow(object):

    def setupUi(self, MainWindow):
        """
        @brief Sets up the user interface components.

        Initializes the main window and creates the layout for user configurations.
        """
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(600, 500)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")

        self.layout = QtWidgets.QVBoxLayout(self.centralwidget)

        # Group boxes
        self.setupGroupBoxes()
        # Save and Load Buttons
        self.setupSaveAndLoadButtons()

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 600, 26))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "User Configuration"))
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

        # Center the window initially
        screen = QtWidgets.QDesktopWidget().screenGeometry()
        screen_width = screen.width()
        screen_height = screen.height()
        window_width = MainWindow.width()
        window_height = MainWindow.height()
        x = (screen_width - window_width) // 2
        y = (screen_height - window_height) // 3
        MainWindow.setGeometry(x, y, window_width, window_height)

    def setupGroupBoxes(self):
        """
        @brief Sets up the group boxes for different configuration sections.
        """
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)

        # Upload Settings group
        self.uploadSettingsGroupBox = QtWidgets.QGroupBox(
            "Upload Settings", self.centralwidget
        )
        self.uploadSettingsLayout = QtWidgets.QGridLayout(self.uploadSettingsGroupBox)

        self.Logger_ID = self.createLabel("Logger ID", font)
        self.lineEdit_Logger_ID = self.createLineEdit(
            "Enter Logger ID (positive integer)"
        )
        self.uploadSettingsLayout.addWidget(self.Logger_ID, 0, 0)
        self.uploadSettingsLayout.addWidget(self.lineEdit_Logger_ID, 0, 1)

        self.Cell_ID = self.createLabel("Cell ID", font)
        self.lineEdit_Cell_ID = self.createLineEdit("Enter Cell ID (positive integer)")
        self.uploadSettingsLayout.addWidget(self.Cell_ID, 1, 0)
        self.uploadSettingsLayout.addWidget(self.lineEdit_Cell_ID, 1, 1)

        self.Upload_Method = self.createLabel("Upload Method", font)
        self.comboBox_Upload_Method = QtWidgets.QComboBox(self.centralwidget)
        self.comboBox_Upload_Method.addItems(["WiFi", "LoRa"])
        self.comboBox_Upload_Method.setCurrentIndex(1)  # Set default to LoRa
        self.comboBox_Upload_Method.currentIndexChanged.connect(self.toggleUploadMethod)
        self.uploadSettingsLayout.addWidget(self.Upload_Method, 2, 0)
        self.uploadSettingsLayout.addWidget(self.comboBox_Upload_Method, 2, 1)

        self.Upload_Interval = self.createLabel("Upload Interval", font)
        self.layout_Upload_Interval = QtWidgets.QHBoxLayout()

        # Upload_Interval: Input fields for Days, Hours, Minutes, Seconds
        self.lineEdit_Days = self.createLineEdit("0")
        self.lineEdit_Days.setFixedWidth(40)
        self.layout_Upload_Interval.addWidget(self.lineEdit_Days)
        self.label_Days = QtWidgets.QLabel("days")
        self.layout_Upload_Interval.addWidget(self.label_Days)

        self.lineEdit_Hours = self.createLineEdit("0")
        self.lineEdit_Hours.setFixedWidth(40)
        self.layout_Upload_Interval.addWidget(self.lineEdit_Hours)
        self.label_Hours = QtWidgets.QLabel("hours")
        self.layout_Upload_Interval.addWidget(self.label_Hours)

        self.lineEdit_Minutes = self.createLineEdit("0")
        self.lineEdit_Minutes.setFixedWidth(40)
        self.layout_Upload_Interval.addWidget(self.lineEdit_Minutes)
        self.label_Minutes = QtWidgets.QLabel("minutes")
        self.layout_Upload_Interval.addWidget(self.label_Minutes)

        self.lineEdit_Seconds = self.createLineEdit("0")
        self.lineEdit_Seconds.setFixedWidth(40)
        self.layout_Upload_Interval.addWidget(self.lineEdit_Seconds)
        self.label_Seconds = QtWidgets.QLabel("seconds")
        self.layout_Upload_Interval.addWidget(self.label_Seconds)

        self.uploadSettingsLayout.addWidget(self.Upload_Interval, 3, 0)
        self.uploadSettingsLayout.addLayout(self.layout_Upload_Interval, 3, 1)

        self.layout.addWidget(self.uploadSettingsGroupBox)

        # Measurement Settings group
        self.measurementSettingsGroupBox = QtWidgets.QGroupBox(
            "Measurement Settings", self.centralwidget
        )
        self.measurementSettingsLayout = QtWidgets.QGridLayout(
            self.measurementSettingsGroupBox
        )

        self.Enabled_Sensors = self.createLabel("Enabled Sensors", font)
        self.checkBox_Voltage = QtWidgets.QCheckBox("Voltage")
        self.checkBox_Current = QtWidgets.QCheckBox("Current")
        self.checkBox_Teros12 = QtWidgets.QCheckBox("Teros12")
        self.checkBox_Teros21 = QtWidgets.QCheckBox("Teros21")
        self.checkBox_BME280 = QtWidgets.QCheckBox("BME280")

        self.measurementSettingsLayout.addWidget(self.Enabled_Sensors, 0, 0)
        self.measurementSettingsLayout.addWidget(self.checkBox_Voltage, 0, 1)
        self.measurementSettingsLayout.addWidget(self.checkBox_Current, 1, 1)
        self.measurementSettingsLayout.addWidget(self.checkBox_Teros12, 2, 1)
        self.measurementSettingsLayout.addWidget(self.checkBox_Teros21, 3, 1)
        self.measurementSettingsLayout.addWidget(self.checkBox_BME280, 4, 1)

        self.Calibration_V_Slope = self.createLabel("Calibration V Slope", font)
        self.lineEdit_V_Slope = self.createLineEdit(
            "Enter Voltage Slope (floating-point)"
        )
        self.measurementSettingsLayout.addWidget(self.Calibration_V_Slope, 5, 0)
        self.measurementSettingsLayout.addWidget(self.lineEdit_V_Slope, 5, 1)

        self.Calibration_V_Offset = self.createLabel("Calibration V Offset", font)
        self.lineEdit_V_Offset = self.createLineEdit(
            "Enter Voltage Offset (floating-point)"
        )
        self.measurementSettingsLayout.addWidget(self.Calibration_V_Offset, 6, 0)
        self.measurementSettingsLayout.addWidget(self.lineEdit_V_Offset, 6, 1)

        self.Calibration_I_Slope = self.createLabel("Calibration I Slope", font)
        self.lineEdit_I_Slope = self.createLineEdit(
            "Enter Current Slope (floating-point)"
        )
        self.measurementSettingsLayout.addWidget(self.Calibration_I_Slope, 7, 0)
        self.measurementSettingsLayout.addWidget(self.lineEdit_I_Slope, 7, 1)

        self.Calibration_I_Offset = self.createLabel("Calibration I Offset", font)
        self.lineEdit_I_Offset = self.createLineEdit(
            "Enter Current Offset (floating-point)"
        )
        self.measurementSettingsLayout.addWidget(self.Calibration_I_Offset, 8, 0)
        self.measurementSettingsLayout.addWidget(self.lineEdit_I_Offset, 8, 1)

        self.layout.addWidget(self.measurementSettingsGroupBox)

        # WiFi Settings group (initially hidden)
        self.wifiGroupBox = QtWidgets.QGroupBox(
            "WiFi Configuration", self.centralwidget
        )
        self.wifiLayout = QtWidgets.QGridLayout(self.wifiGroupBox)

        self.WiFi_SSID = self.createLabel("WiFi SSID", font)
        self.lineEdit_WiFi_SSID = self.createLineEdit("Enter WiFi SSID")
        self.wifiLayout.addWidget(self.WiFi_SSID, 0, 0)
        self.wifiLayout.addWidget(self.lineEdit_WiFi_SSID, 0, 1)

        self.WiFi_Password = self.createLabel("WiFi Password", font)
        self.lineEdit_WiFi_Password = self.createLineEdit("Enter WiFi Password")
        self.lineEdit_WiFi_Password.setEchoMode(QtWidgets.QLineEdit.Password)
        self.wifiLayout.addWidget(self.WiFi_Password, 1, 0)
        self.wifiLayout.addWidget(self.lineEdit_WiFi_Password, 1, 1)

        self.API_Endpoint_URL = self.createLabel("API Endpoint URL", font)
        self.lineEdit_API_Endpoint_URL = self.createLineEdit(
            "Enter API Endpoint URL (start with http:// or https://)"
        )
        self.wifiLayout.addWidget(self.API_Endpoint_URL, 2, 0)
        self.wifiLayout.addWidget(self.lineEdit_API_Endpoint_URL, 2, 1)

        self.API_Port = self.createLabel("API Port", font)
        self.lineEdit_API_Port = self.createLineEdit("Enter API Port (integer)")
        self.wifiLayout.addWidget(self.API_Port, 3, 0)
        self.wifiLayout.addWidget(self.lineEdit_API_Port, 3, 1)

        # Ensure consistent size for wifiGroupBox
        self.wifiGroupBox.setFixedHeight(self.wifiGroupBox.minimumSizeHint().height())
        self.layout.addWidget(self.wifiGroupBox)

        # Show or hide WiFi settings based on upload method
        self.toggleUploadMethod()

    def toggleUploadMethod(self):
        """
        @brief Shows or hides WiFi settings based on the upload method selected.
        """
        if self.comboBox_Upload_Method.currentText() == "WiFi":
            self.showWiFiSettings()
        else:
            self.hideWiFiSettings()

    def showWiFiSettings(self):
        """
        @brief Displays the WiFi configuration settings.
        """
        self.lineEdit_WiFi_SSID.setEnabled(True)
        self.lineEdit_WiFi_Password.setEnabled(True)
        self.lineEdit_API_Endpoint_URL.setEnabled(True)
        self.lineEdit_API_Port.setEnabled(True)
        self.lineEdit_WiFi_SSID.show()
        self.lineEdit_WiFi_Password.show()
        self.lineEdit_API_Endpoint_URL.show()
        self.lineEdit_API_Port.show()
        # set default values for API URL & PORT
        self.lineEdit_API_Endpoint_URL.setText("https://dirtviz.jlab.ucsc.edu/api/")
        self.lineEdit_API_Port.setText("443")

    def hideWiFiSettings(self):
        """
        @brief Hides the WiFi configuration settings.
        """
        self.lineEdit_WiFi_SSID.setEnabled(False)
        self.lineEdit_WiFi_Password.setEnabled(False)
        self.lineEdit_API_Endpoint_URL.setEnabled(False)
        self.lineEdit_API_Port.setEnabled(False)
        self.lineEdit_WiFi_SSID.hide()
        self.lineEdit_WiFi_Password.hide()
        self.lineEdit_API_Endpoint_URL.hide()
        self.lineEdit_API_Port.hide()

    def createLabel(self, text, font):
        """
        @brief Creates a label with the specified text and font.

        @param text Text for the label.
        @param font Font settings for the label.
        @return QLabel instance
        """
        label = QtWidgets.QLabel(text)
        label.setFont(font)
        return label

    def createLineEdit(self, placeholder):
        """
        @brief Creates a QLineEdit with a placeholder.

        @param placeholder Placeholder text for the QLineEdit.
        @return QLineEdit instance
        """
        lineEdit = QtWidgets.QLineEdit(self.centralwidget)
        lineEdit.setPlaceholderText(placeholder)
        return lineEdit

    def setupSaveAndLoadButtons(self):
        """
        @brief Creates and configures the save and load buttons.
        """
        # Create a grid layout for precise placement
        button_layout = QtWidgets.QGridLayout()

        # Load button
        self.loadButton = QtWidgets.QPushButton("Load", self.centralwidget)
        self.loadButton.setFixedSize(100, 30)
        self.loadButton.clicked.connect(self.loadConfiguration)
        button_layout.addWidget(self.loadButton, 1, 0)  # Row 0, Column 0

        # Save button
        self.saveButton = QtWidgets.QPushButton("Save", self.centralwidget)
        self.saveButton.setFixedSize(100, 30)
        self.saveButton.clicked.connect(self.saveConfiguration)
        button_layout.addWidget(self.saveButton, 1, 1)  # Row 0, Column 1

        # Send Configuration button
        self.saveConfigurationButton = QtWidgets.QPushButton(
            "Send Configuration", self.centralwidget
        )
        self.saveConfigurationButton.setFixedSize(300, 30)
        self.saveConfigurationButton.clicked.connect(
            lambda: self.saveConfiguration(flag="send")
        )
        button_layout.addWidget(self.saveConfigurationButton, 1, 2)  # Row 0, Column 2

        # Load current Configuration button
        self.loadCurrentConfigButton = QtWidgets.QPushButton(
            "Load current Configuration", self.centralwidget
        )
        self.loadCurrentConfigButton.setFixedSize(300, 30)
        self.loadCurrentConfigButton.clicked.connect(
            lambda: self.loadConfiguration(flag="loadCurrent")
        )
        button_layout.addWidget(self.loadCurrentConfigButton, 0, 2)  # Row 1, Column 2

        # Add the grid layout to the main layout
        self.layout.addLayout(button_layout)

    def saveConfiguration(self, flag: str):
        """
        @brief Validates inputs, encodes the configuration, and sends it via UART.
        """
        try:
            logger_id = self.validateUInt(self.lineEdit_Logger_ID.text(), "Logger ID")
            cell_id = self.validateUInt(self.lineEdit_Cell_ID.text(), "Cell ID")
            upload_method = self.comboBox_Upload_Method.currentText()

            # Calculate upload interval in seconds
            days = (
                0
                if self.lineEdit_Days.text() == ""
                else self.validateUInt(self.lineEdit_Days.text(), "Days")
            )
            hours = (
                0
                if self.lineEdit_Hours.text() == ""
                else self.validateUInt(self.lineEdit_Hours.text(), "Hours")
            )
            minutes = (
                0
                if self.lineEdit_Minutes.text() == ""
                else self.validateUInt(self.lineEdit_Minutes.text(), "Minutes")
            )
            seconds = (
                0
                if self.lineEdit_Seconds.text() == ""
                else self.validateUInt(self.lineEdit_Seconds.text(), "Seconds")
            )
            upload_interval = days * 86400 + hours * 3600 + minutes * 60 + seconds

            # Check if the user entered time in upload interval
            if upload_interval == 0:
                raise ValueError('You must Enter preferred time in "upload interval".')

            # Check if the user selected at least one sensor option
            if not (
                self.checkBox_Voltage.isChecked()
                or self.checkBox_Current.isChecked()
                or self.checkBox_Teros12.isChecked()
                or self.checkBox_Teros21.isChecked()
                or self.checkBox_BME280.isChecked()
            ):
                raise ValueError("You must choose at least one sensor.")

            # Convert enabled sensors into a list and filter out empty strings
            enabled_sensors = [
                sensor
                for sensor in [
                    "Voltage" if self.checkBox_Voltage.isChecked() else "",
                    "Current" if self.checkBox_Current.isChecked() else "",
                    "Teros12" if self.checkBox_Teros12.isChecked() else "",
                    "Teros21" if self.checkBox_Teros21.isChecked() else "",
                    "BME280" if self.checkBox_BME280.isChecked() else "",
                ]
                if sensor
            ]

            # Checked sensors to be saved in json
            enabled_sensors_json = {
                "Voltage": self.checkBox_Voltage.isChecked(),
                "Current": self.checkBox_Current.isChecked(),
                "Teros12": self.checkBox_Teros12.isChecked(),
                "Teros21": self.checkBox_Teros21.isChecked(),
                "BME280": self.checkBox_BME280.isChecked(),
            }
            calibration_v_slope = self.validateFloat(
                self.lineEdit_V_Slope.text(), "Calibration V Slope"
            )
            calibration_v_offset = self.validateFloat(
                self.lineEdit_V_Offset.text(), "Calibration V Offset"
            )
            calibration_i_slope = self.validateFloat(
                self.lineEdit_I_Slope.text(), "Calibration I Slope"
            )
            calibration_i_offset = self.validateFloat(
                self.lineEdit_I_Offset.text(), "Calibration I Offset"
            )

            # Add WiFi settings if WiFi is selected as the upload method
            if upload_method == "WiFi":
                wifi_ssid = self.lineEdit_WiFi_SSID.text()
                wifi_password = self.lineEdit_WiFi_Password.text()
                api_endpoint_url = self.validateURL(
                    self.lineEdit_API_Endpoint_URL.text(), "API Endpoint URL"
                )
                api_port = self.validateUInt(self.lineEdit_API_Port.text(), "API Port")
            else:
                wifi_ssid = ""
                wifi_password = ""
                api_endpoint_url = ""
                api_port = 0

            # Validate user input on case of WiFi
            if upload_method == "WiFi":
                if not self.lineEdit_WiFi_SSID.text():
                    raise ValueError("WiFi SSID cannot be empty.")

            # Construct the configuration dictionary to be saved in json file
            configuration = {
                "Logger ID": logger_id,
                "Cell ID": cell_id,
                "Upload Method": upload_method,
                "Days": days,
                "Hours": hours,
                "Minutes": minutes,
                "Seconds": seconds,
                "Enabled Sensors": enabled_sensors_json,
                "Calibration V Slope": calibration_v_slope,
                "Calibration V Offset": calibration_v_offset,
                "Calibration I Slope": calibration_i_slope,
                "Calibration I Offset": calibration_i_offset,
                "WiFi SSID": wifi_ssid,
                "WiFi Password": wifi_password,
                "API Endpoint URL": api_endpoint_url,
                "API Port": api_port,
            }

            # Check whether the user wants to send or just to save
            if flag == "send":
                # Construct the configuration dictionary to be encoded
                encoded_data = encode_user_configuration(
                    int(logger_id),
                    int(cell_id),
                    upload_method,
                    int(upload_interval),
                    enabled_sensors,
                    float(calibration_v_slope),
                    float(calibration_v_offset),
                    float(calibration_i_slope),
                    float(calibration_i_offset),
                    wifi_ssid,
                    wifi_password,
                    api_endpoint_url,
                    int(api_port),
                )
                # Send the encoded configuration via UART
                success = self.sendToUART(encoded_data)
                if not success:
                    return  # Don't save if sending failed

                print("------------------------------------------")
                print(encoded_data)
                print("------------------------------------------")
            try:
                # Ensure the 'Load' directory exists
                load_dir = os.path.join(os.path.dirname(__file__), "Load")
                os.makedirs(load_dir, exist_ok=True)
                # Save configuration as JSON
                config_path = os.path.join(load_dir, f"cell_{cell_id}_.json")
                with open(config_path, "w") as json_file:
                    json.dump(configuration, json_file, indent=4)

                QtWidgets.QMessageBox.information(
                    self.centralwidget, "Success", "Configurations saved successfully."
                )
            except Exception as e:
                QtWidgets.QMessageBox.critical(
                    self.centralwidget, "Error", f"Failed to save configurations: {e}"
                )

            # Print success message in case of saving the config or saving and sending the config
            if flag == "send":
                print(
                    f"Configuration saved and sent to STM32 successfully! Backup JSON file: {'cell_'+ str(cell_id) + '_' }"
                )
            else:
                print(
                    f"Configuration saved successfully! Backup JSON file: {'cell_'+ str(cell_id) + '_' }"
                )

        except ValueError as e:
            # Show error message if validation fails
            QtWidgets.QMessageBox.critical(self.centralwidget, "Error", str(e))

    def loadConfiguration(self, flag: str):
        """
        @brief Loads configuration from a selected JSON file and fills the input fields.
        """
        # Load current configuration in STM32 and display it on the GUI
        if flag == "loadCurrent":
            success, encoded_data = self.receiveFromUART()
            if not success:
                return
            decoded_data = decode_user_configuration(encoded_data)
            print(decoded_data)

            # Update GUI elements with decoded data
            self.lineEdit_Logger_ID.setText(str(decoded_data["loggerId"]))
            self.lineEdit_Cell_ID.setText(str(decoded_data["cellId"]))
            self.comboBox_Upload_Method.setCurrentText(decoded_data["UploadMethod"])

            # Calculate upload interval and update GUI fields
            upload_interval = decoded_data["UploadInterval"]
            days = upload_interval // 86400
            remaining_seconds = upload_interval % 86400
            hours = remaining_seconds // 3600
            remaining_seconds %= 3600
            minutes = remaining_seconds // 60
            seconds = remaining_seconds % 60
            self.lineEdit_Days.setText(str(days))
            self.lineEdit_Hours.setText(str(hours))
            self.lineEdit_Minutes.setText(str(minutes))
            self.lineEdit_Seconds.setText(str(seconds))

            # Update sensor checkboxes
            self.checkBox_Voltage.setChecked(False)
            self.checkBox_Current.setChecked(False)
            self.checkBox_Teros12.setChecked(False)
            self.checkBox_Teros21.setChecked(False)
            self.checkBox_BME280.setChecked(False)
            for sensor in decoded_data["enabledSensors"]:
                if sensor == "Voltage":
                    self.checkBox_Voltage.setChecked(True)
                elif sensor == "Current":
                    self.checkBox_Current.setChecked(True)
                elif sensor == "Teros12":
                    self.checkBox_Teros12.setChecked(True)
                elif sensor == "Teros21":
                    self.checkBox_Teros21.setChecked(True)
                elif sensor == "BME280":
                    self.checkBox_BME280.setChecked(True)
            # Fill calibration fields
            self.lineEdit_V_Slope.setText(str(decoded_data["VoltageSlope"]))
            self.lineEdit_V_Offset.setText(str(decoded_data["VoltageOffset"]))
            self.lineEdit_I_Slope.setText(str(decoded_data["CurrentSlope"]))
            self.lineEdit_I_Offset.setText(str(decoded_data["CurrentOffset"]))

            # Fill WiFi settings if upload method is WiFi
            if decoded_data["UploadMethod"] == "WiFi":
                self.lineEdit_WiFi_SSID.setText(decoded_data["WiFiSSID"])
                self.lineEdit_WiFi_Password.setText(decoded_data["WiFiPassword"])
                self.lineEdit_API_Endpoint_URL.setText(decoded_data["APIEndpointURL"])
                self.lineEdit_API_Port.setText(str(decoded_data["APIEndpointPort"]))
            else:
                # Clear WiFi fields if upload method is not WiFi
                self.lineEdit_WiFi_SSID.clear()
                self.lineEdit_WiFi_Password.clear()
                self.lineEdit_API_Endpoint_URL.clear()
                self.lineEdit_API_Port.clear()

            QtWidgets.QMessageBox.information(
                self.centralwidget,
                "Success",
                "Configuration loaded successfully From FRAM.",
            )
            return
        # Load configuration from JSON file
        else:
            # Open a file dialog to select the JSON file
            options = QtWidgets.QFileDialog.Options()
            file_path, _ = QtWidgets.QFileDialog.getOpenFileName(
                self.centralwidget,
                "Select Configuration File",
                "",
                "JSON Files (*.json)",
                options=options,
            )

        if file_path:
            try:
                with open(file_path, "r") as json_file:
                    config = json.load(json_file)

                # Fill the GUI fields with the loaded configuration
                self.lineEdit_Logger_ID.setText(str(config.get("Logger ID", "")))
                self.lineEdit_Cell_ID.setText(str(config.get("Cell ID", "")))
                self.comboBox_Upload_Method.setCurrentText(
                    config.get("Upload Method", "WiFi")
                )

                # Fill the upload interval fields
                self.lineEdit_Days.setText(str(config.get("Days", 0)))
                self.lineEdit_Hours.setText(str(config.get("Hours", 0)))
                self.lineEdit_Minutes.setText(str(config.get("Minutes", 0)))
                self.lineEdit_Seconds.setText(str(config.get("Seconds", 0)))

                # Fill sensor checkboxes
                enabled_sensors = config.get("Enabled Sensors", {})
                self.checkBox_Voltage.setChecked(enabled_sensors.get("Voltage", False))
                self.checkBox_Current.setChecked(enabled_sensors.get("Current", False))
                self.checkBox_Teros12.setChecked(enabled_sensors.get("Teros12", False))
                self.checkBox_Teros21.setChecked(enabled_sensors.get("Teros21", False))
                self.checkBox_BME280.setChecked(enabled_sensors.get("BME280", False))

                # Fill calibration fields
                self.lineEdit_V_Slope.setText(
                    str(config.get("Calibration V Slope", ""))
                )
                self.lineEdit_V_Offset.setText(
                    str(config.get("Calibration V Offset", ""))
                )
                self.lineEdit_I_Slope.setText(
                    str(config.get("Calibration I Slope", ""))
                )
                self.lineEdit_I_Offset.setText(
                    str(config.get("Calibration I Offset", ""))
                )

                # Fill WiFi settings if upload method is WiFi
                if config.get("Upload Method") == "WiFi":
                    self.lineEdit_WiFi_SSID.setText(config.get("WiFi SSID", ""))
                    self.lineEdit_WiFi_Password.setText(config.get("WiFi Password", ""))
                    self.lineEdit_API_Endpoint_URL.setText(
                        config.get("API Endpoint URL", "")
                    )
                    self.lineEdit_API_Port.setText(str(config.get("API Port", "")))
                else:
                    # Clear the WiFi fields if the method is not WiFi
                    self.lineEdit_WiFi_SSID.clear()
                    self.lineEdit_WiFi_Password.clear()
                    self.lineEdit_API_Endpoint_URL.clear()
                    self.lineEdit_API_Port.clear()

                QtWidgets.QMessageBox.information(
                    self.centralwidget, "Success", "Configuration loaded successfully."
                )

            except Exception as e:
                QtWidgets.QMessageBox.critical(
                    self.centralwidget, "Error", f"Failed to load configuration: {e}"
                )

    def sendToUART(self, encoded_data):
        """
        @brief Sends the encoded configuration data via UART.

        @param data Encoded configuration data.
        """
        ser = None
        try:
            # List available ports with descriptions
            ports = serial.tools.list_ports.comports()
            available_ports = [f"{port.device} - {port.description}" for port in ports]

            if not available_ports:
                QtWidgets.QMessageBox.critical(
                    self.centralwidget, "Error", "No serial ports available."
                )
                return False

            # Ask the user to select a port
            selected_port, ok = QInputDialog.getItem(
                self.centralwidget,
                "Select Port",
                "Available Serial Ports:",
                available_ports,
                0,
                False,
            )

            if not ok or not selected_port:
                QtWidgets.QMessageBox.critical(
                    self.centralwidget, "Error", "No port selected."
                )
                return False

            # Extract the port name
            port_name = selected_port.split(" ")[0]

            # Open the serial port
            ser = serial.Serial(port=port_name, baudrate=115200, timeout=20)
            # Step 0: Send 1 indicating sending new config to be stored
            ser.flush()
            ser.write(bytes([1]))
            print(f"Sending: {bytes([1])}")
            # Step 1: Send the length of the encoded data (2 bytes)
            data_length = len(encoded_data)
            ser.write(
                data_length.to_bytes(2, byteorder="big")
            )  # Send length as 2-byte big-endian integer
            # Step 2: Send the encoded data
            ser.write(encoded_data)
            print(
                "________________________________________________________________________"
            )
            print(f"length: {data_length}")
            print(f"{encoded_data}")

            # Step 3: Wait for acknowledgment ("ACK")
            ack = ser.read(3)  # Read 3 bytes (assuming "ACK" is 3 bytes)
            print(ack)
            print(
                "________________________________________________________________________"
            )
            if ack == b"ACK":
                QtWidgets.QMessageBox.information(
                    self.centralwidget, "Success", "Received ACK from STM32"
                )
            else:
                QtWidgets.QMessageBox.critical(
                    self.centralwidget, "UART Error", "No acknowledgment received"
                )
                return False

            # Step 4: After ACK, read back the same data from STM32
            received_data_length = int.from_bytes(
                ser.read(2), byteorder="big"
            )  # Read the length of received data
            print(
                "________________________________________________________________________"
            )
            print(f"length: {received_data_length}")
            print(
                "________________________________________________________________________"
            )
            print(
                "________________________________________________________________________"
            )

            received_data = ser.read(
                received_data_length
            )  # Read the received data based on the length
            print(f"Received from STM32: {received_data}")
            print(
                "________________________________________________________________________"
            )

            # Step 5: Display the received data to confirm it's the same
            if received_data == encoded_data:
                QtWidgets.QMessageBox.information(
                    self.centralwidget,
                    "Success",
                    "Data received matches the sent data.",
                )
            else:
                QtWidgets.QMessageBox.critical(
                    self.centralwidget,
                    "Error",
                    "Received data does not match sent data.",
                )

            return True

        except serial.SerialException as e:
            QtWidgets.QMessageBox.critical(
                self.centralwidget, "UART Error", f"Failed to send data: {e}"
            )
            return False

        finally:
            if ser is not None:
                ser.close()

    def receiveFromUART(self):
        """
        @brief Receives the current encoded configuration data via UART.

        @param void.
        @return (success, data): A tuple containing success status and decoded data or an error message.
        """
        ser = None
        try:
            # List available ports with descriptions
            ports = serial.tools.list_ports.comports()
            available_ports = [f"{port.device} - {port.description}" for port in ports]

            if not available_ports:
                QtWidgets.QMessageBox.critical(
                    self.centralwidget, "Error", "No serial ports available."
                )
                return False, "No serial ports available."

            # Ask the user to select a port
            selected_port, ok = QInputDialog.getItem(
                self.centralwidget,
                "Select Port",
                "Available Serial Ports:",
                available_ports,
                0,
                False,
            )

            if not ok or not selected_port:
                QtWidgets.QMessageBox.critical(
                    self.centralwidget, "Error", "No port selected."
                )
                return False, "No port selected."

            # Extract the port name
            port_name = selected_port.split(" ")[0]

            # Open the serial port
            ser = serial.Serial(port=port_name, baudrate=115200, timeout=2)
            # Step 0: Send 2 indicating loading the current configurations from the FRAM
            ser.write(bytes([2]))
            print(f"Sending: {bytes([2])}")
            # Step 1: Wait for acknowledgment ("ACK")
            ack = ser.read(3)  # Read 3 bytes (assuming "ACK" is 3 bytes)
            print(ack)
            if ack == b"ACK":
                QtWidgets.QMessageBox.information(
                    self.centralwidget, "Success", "Received ACK from STM32"
                )
            else:
                QtWidgets.QMessageBox.critical(
                    self.centralwidget, "UART Error", "No acknowledgment received"
                )
                return False, "No acknowledgment received from STM32."

            # Step 2: After ACK, read data from STM32
            received_data_length = int.from_bytes(
                ser.read(2), byteorder="big"
            )  # Read the length of received data
            print(
                "________________________________________________________________________"
            )
            print(f"length: {received_data_length}")
            print(
                "________________________________________________________________________"
            )
            print(
                "________________________________________________________________________"
            )

            received_data = ser.read(
                received_data_length
            )  # Read the received data based on the length
            print(f"Received from STM32: {received_data}")
            print(
                "________________________________________________________________________"
            )
            return True, received_data

        except serial.SerialException as e:
            QtWidgets.QMessageBox.critical(
                self.centralwidget, "UART Error", f"Failed to send data: {e}"
            )
            return False, f"Failed to send or receive data: {e}"

        finally:
            if ser is not None:
                ser.close()

    def validateURL(self, value, field_name):
        """
        @brief Validates that the input is a valid URL.
        @param value The input value to validate.
        @param field_name The name of the field being validated.
        @return The validated URL string.
        @throws ValueError if the input is invalid.
        """
        url_pattern = re.compile(
            r"^(https?|ftp):\/\/"  # http:// or https:// or ftp://
            r"([a-zA-Z0-9_-]+(?:(?:\.[a-zA-Z0-9_-]+)+))"  # domain
            r"(:[0-9]{1,5})?"  # port (?: optional)
            r"(\/.*)?$"  # path (?: optional)
        )
        if not url_pattern.match(value):
            raise ValueError(f"Invalid {field_name}. Must be a valid URL.")
        return value

    def validateUInt(self, value, name):
        """
        @brief Validates that the input is a positive integer.

        @param value The input value to validate.
        @param name The name of the parameter (for error messages).
        @return Validated unsigned integer value.
        """
        if not value.isdigit() or int(value) < 0:
            raise ValueError(f"{name} must be a positive integer.")
        return int(value)

    def validateInt(self, value, name):
        """
        @brief Validates that the input is an integer.

        @param value The input value to validate.
        @param name The name of the parameter (for error messages).
        @return Validated integer value.
        """
        try:
            return int(value)
        except ValueError:
            raise ValueError(f"{name} must be an integer.")

    def validateFloat(self, value, name):
        """
        @brief Validates that the input is a float.

        @param value The input value to validate.
        @param name The name of the parameter (for error messages).
        @return Validated float value.
        """
        try:
            return float(value)
        except ValueError:
            raise ValueError(f"{name} must be a floating-point number.")


def main():
    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    ui = Ui_MainWindow()
    ui.setupUi(MainWindow)
    MainWindow.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



"""To test connection to Soil Power Sensor over serial

Stephen Taylor 5/20/2024
"""

import serial


class SerialPing:
    """Simple serial ping utility"""

    def __init__(self, port, baudrate=115200, xonxoff=True):
        """Constructor

        Initializes serial connection.

        Parameters
        ----------
        port : str
            Serial port name (e.g., COM1, /dev/ttyUSB0)
        baudrate : int, optional
            Baud rate for serial communication (default is 115200, STM32 functions at 115200)
        xonxoff  : bool, optional
            Flow control (default is on)
        """
        self.ser = serial.Serial(port, baudrate=baudrate, xonxoff=xonxoff, timeout=1)
        # Print serial port settings
        print("Serial Port Settings:")
        print("Port:", self.ser.port)
        print("Baudrate:", self.ser.baudrate)
        print("Byte size:", self.ser.bytesize)
        print("Parity:", self.ser.parity)
        print("Stop bits:", self.ser.stopbits)
        print("Timeout:", self.ser.timeout)
        print("Xon/Xoff:", self.ser.xonxoff)
        print("Rts/cts:", self.ser.rtscts)
        print("Dsr/dtr:", self.ser.dsrdtr)

    def ping(self):
        """Ping the device and return the response"""
        # pdb.set_trace()
        self.ser.write(b"check\n")  # Send ping command
        response = self.ser.readline()  # Read response
        return response

    def close(self):
        """Close serial connection"""
        self.ser.close()


# Example usage
if __name__ == "__main__":
    # Replace 'COM1' with the appropriate serial port on your system
    port = "COM14"

    # Create SerialPing object
    serial_ping = SerialPing(port)

    try:
        # Ping the device
        response = serial_ping.ping()
        print("Response:", response)

    finally:
        # Close serial connection
        serial_ping.close()



#!/usr/bin/env python

"""Data recorder for Soil Power Sensor

The recorder controls the SPS firmware and a Keithley 2400 Source
Measurement Unit (SMU). Ideally the script should work with any microcontrollre
flashed with the firmware and any source measurement unit that supports
Standard Commands for Programable Instruments (SCPI). The units listed are the
ones that the script was developed and tested on. It allows to step through
a range of output voltages on the Keithley and measure the voltage and current
from both the SMU and the Soil Power Sensor (SPS).
"""

import time
import socket
import serial
from typing import Tuple
from tqdm import tqdm
from ..proto import decode_measurement


class SerialController:
    """Generic serial controller that will open and close serial connections"""

    # Serial port
    ser = None

    def __init__(self, port, baudrate=115200, xonxoff=False):
        """Constructor

        Initialises connection to serial port.

        Parameters
        ----------
        port : str
            Serial port of device
        baudrate : int, optional
            Baud rate for serial communication (default is 115200, STM32 functions at 115200)
        xonxoff  : bool, optional
            Flow control (default is on)
        """

        self.ser = serial.Serial(port, baudrate=baudrate, xonxoff=xonxoff)

    def __del__(self):
        """Destructor

        Closes connection to serial port.
        """

        self.ser.close()


class LANController:
    """Generic LAN controller that will open and close LAN connections"""

    # Socket
    sock = None

    def __init__(self, host, port):
        """Constructor

        Initialises connection to LAN device.

        Parameters
        ----------
        host : str
            IP address or hostname of the device
        port : int
            Port number of the device
        """

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((host, port))

    def __del__(self):
        """Destructor

        Closes connection to LAN device.
        """

        self.sock.close()


class SoilPowerSensorController(SerialController):
    """Controller used to read values from the SPS"""

    def __init__(self, port):
        """Constructor

        Opens serial connection and checks functionality

        Parameters
        ----------
        port : str
            Serial port of device
        """
        super().__init__(port)
        self.check()

    def get_power(self) -> Tuple[float, float]:
        """Measure voltage from SPS

        Returns
        -------
        tuple[float, float
            voltage, current
        """

        self.ser.write(b"0")  # send a command to the SPS to send a power measurment

        # read a single byte for the length
        resp_len_bytes = self.ser.read()

        resp_len = int.from_bytes(resp_len_bytes)

        reply = self.ser.read(resp_len)  # read said measurment

        meas_dict = decode_measurement(reply)  # decode using protobuf

        voltage_value = meas_dict["data"]["voltage"]
        current_value = meas_dict["data"]["current"]

        return float(voltage_value), float(current_value)

    def check(self):
        """Performs a check of the connection to the board

        Raises
        ------
        RuntimeError
            Checks that SPS replies "ok" when sent "check"
        """

        # needed sleep to get write to work
        # possibly due to linux usb stack initialized or mcu waiting to startup
        time.sleep(1)
        self.ser.write(b"check\n")

        reply = self.ser.readline()
        # reply = reply.decode()
        # reply = reply.strip("\r\n")

        if reply != b"ok\n":
            raise RuntimeError(f"SPS check failed. Reply received: {reply}")


class SMUSerialController(SerialController):
    """Controller for the Keithley 2400 SMU used to supply known voltage to the
    SPS

    Uses SCPI (Standard Control of Programmable Instruments) to control the SMU
    through its RS232 port. Written for the Keithley 2400 SMU, but should work
    for any other SMU that uses SCPI.
    """

    class VoltageIterator:
        """VoltageIterator Class

        Implements a iterator for looping through voltage output values
        """

        def __init__(self, ser, start, stop, step):
            """Constructor

            Parameters
            ----------
            ser : serial.Serial
                Initialised serial connection
            start : float
                Starting voltage
            stop : float
                End voltage
            step : float
                Voltage step
            """

            self.ser = ser
            self.start = start
            self.stop = stop
            self.step = step

        def __iter__(self):
            """Iterator

            Sets current value to start
            """

            self.v = None
            self.ser.write(b":OUTP ON\n")
            return self

        def __next__(self):
            """Next

            Steps to next voltage level, stopping once stop is reached

            Raises
            ------
            StopIteration
                When the next step exceeds the stop level
            """

            if self.v is None:
                return self.set_voltage(self.start)

            v_next = self.v + self.step

            if v_next <= self.stop:
                return self.set_voltage(v_next)
            else:
                raise StopIteration

        def set_voltage(self, v):
            """Sets the voltage output"""

            self.v = v
            cmd = f":SOUR:VOLT:LEV {v}\n"
            self.ser.write(bytes(cmd, "ascii"))
            return self.v

    def __init__(self, port, source_mode):
        """Constructor

        Opens serial port, restore to known defaults

        Parameters
        ----------
        port : str
            Serial port of device
        """

        super().__init__(port)
        # Reset settings
        self.ser.write(b"*RST\n")
        # Voltage source
        self.ser.write(b":SOUR:FUNC VOLT\n")
        self.ser.write(b":SOUR:VOLT:MODE FIXED\n")
        # 1mA compliance
        self.ser.write(b":SENS:CURR:PROT 10e-3\n")
        # Sensing functions
        self.ser.write(b":SENS:CURR:RANGE:AUTO ON\n")
        self.ser.write(b":SENS:FUNC:OFF:ALL\n")
        self.ser.write(b':SENS:FUNC:ON "VOLT"\n')
        self.ser.write(b':SENS:FUNC:ON "CURR"\n')

    def __del__(self):
        """Destructor

        Turns off output
        """

        self.ser.write(b":OUTP OFF\n")

    def vrange(self, start, stop, step) -> VoltageIterator:
        """Gets iterator to range of voltages

        Parameters
        ----------
        start : float
            Starting voltage
        stop : float
            End voltage
        step : float
            Voltage step
        """

        return self.VoltageIterator(self.ser, start, stop, step)

    def get_voltage(self) -> float:
        """Measure voltage supplied to the SPS from SMU

        Returns
        -------
        float
            Measured voltage
        """

        self.ser.write(b":FORM:ELEM VOLT\n")
        self.ser.write(b":READ?\n")
        reply = self.ser.readline().decode()
        reply = reply.strip("\r")
        return float(reply)

    def get_current(self) -> float:
        """Measure current supplied to the SPS from SMU

        Returns
        -------
        float
            Measured current
        """

        self.ser.write(
            b":FORM:ELEM CURR\n"
        )  # replace with serial.write with socket.write commands, std library aviable, lots of example code
        self.ser.write(b":READ?\n")
        reply = self.ser.readline().decode()
        reply = reply.strip("\r")
        return float(reply)


class SMULANController(LANController):
    """Controller for the Keithley 2400 SMU used to supply known voltage to the
    SPS

    Uses SCPI (Standard Control of Programmable Instruments) to control the SMU
    through its RS232 port. Written for the Keithley 2400 SMU, but should work
    for any other SMU that uses SCPI.
    """

    class VoltageIterator:
        """VoltageIterator Class

        Implements a iterator for looping through voltage output values
        """

        def __init__(self, sock, start, stop, step):
            """Constructor

            Parameters
            ----------
            sock : serial.Socket
                Initialised socket connection
            start : float
                Starting voltage
            stop : float
                End voltage
            step : float
                Voltage step
            """

            self.sock = sock
            self.start = start
            self.stop = stop
            self.step = step

        def __iter__(self):
            """Iterator

            Sets current value to start
            """

            self.v = None
            self.sock.sendall(b":OUTP ON\n")
            return self

        def __next__(self):
            """Next

            Steps to next voltage level, stopping once stop is reached

            Raises
            ------
            StopIteration
                When the next step exceeds the stop level
            """

            if self.v is None:
                return self.set_voltage(self.start)

            v_next = self.v + self.step

            if v_next <= self.stop:
                return self.set_voltage(v_next)
            else:
                raise StopIteration

        def __len__(self):
            """Len

            The number of measurements points
            """
            return int((self.stop - self.start) / self.step) + 1

        def set_voltage(self, v):
            """Sets the voltage output"""

            self.v = v
            cmd = f":SOUR:VOLT:LEV {v}\n"
            self.sock.sendall(bytes(cmd, "ascii"))
            return self.v

    class CurrentIterator:
        """CurrentIterator Class

        Implements a iterator for looping through current output values
        """

        def __init__(self, sock, start, stop, step):
            """Constructor

            Parameters
            ----------
            sock : serial.Socket
                Initialised socket connection
            start : float
                Starting current
            stop : float
                End current
            step : float
                Current step
            """
            self.sock = sock
            self.start = start
            self.stop = stop
            self.step = step

        def __iter__(self):
            """Iterator

            Sets current value to start
            """

            self.i = None
            self.sock.sendall(b":OUTP ON\n")
            return self

        def __next__(self):
            """Next

            Steps to next voltage level, stopping once stop is reached

            Raises
            ------
            StopIteration
                When the next step exceeds the stop level
            """

            if self.i is None:
                return self.set_current(self.start)
            i_next = self.i + self.step
            if i_next <= self.stop:
                return self.set_current(i_next)
            else:
                raise StopIteration

        def __len__(self):
            """Len

            The number of measurements points
            """
            return int((self.stop - self.start) / self.step) + 1

        def set_current(self, i):
            """Sets the current output"""

            self.i = i
            cmd = f":SOUR:CURR:LEV {i}\n"
            self.sock.sendall(bytes(cmd, "ascii"))
            return self.i

    def __init__(self, host, port):
        """Constructor

        Opens LAN connection and sets initial SMU configurations.

        Parameters
        ----------
        host : str
            IP address or hostname of the SMU
        port : int
            Port number used for the LAN connection
        """

        super().__init__(host, port)

    def setup_voltage(self):
        """Configures smu for sourcing voltage"""

        self.sock.sendall(b"*RST\n")
        # Voltage source
        self.sock.sendall(b":SOUR:FUNC VOLT\n")
        self.sock.sendall(b":SOUR:VOLT:MODE FIXED\n")
        # 1mA compliance
        self.sock.sendall(b":SENS:CURR:PROT 10e-3\n")
        # Sensing functions
        self.sock.sendall(b":SENS:CURR:RANGE:AUTO ON\n")
        self.sock.sendall(b":SENS:FUNC:OFF:ALL\n")
        self.sock.sendall(b':SENS:FUNC:ON "VOLT"\n')
        self.sock.sendall(b':SENS:FUNC:ON "CURR"\n')

    def setup_current(self):
        """Configured smu for sourcing current"""

        self.sock.sendall(b"*RST\n")
        # Current source
        self.sock.sendall(b":SOUR:FUNC CURR\n")
        self.sock.sendall(b":SOUR:CURR:MODE FIXED\n")
        # 1V compliance
        self.sock.sendall(b":SENSE:CURR:PROT 1\n")
        # Sensing functions
        self.sock.sendall(b":SENS:VOLT:RANGE:AUTO ON\n")
        self.sock.sendall(b":SENS:FUNC:OFF:ALL\n")
        self.sock.sendall(b':SENS:FUNC:ON "CURR"\n')
        self.sock.sendall(b':SENS:FUNC:ON "VOLT"\n')

    def __del__(self):
        """Destructor

        Turns off output
        """

        self.sock.sendall(b":OUTP OFF\n")

    def vrange(self, start, stop, step) -> VoltageIterator:
        """Gets iterator to range of voltages

        Parameters
        ----------
        start : float
            Starting voltage
        stop : float
            End voltage
        step : float
            Voltage step
        """

        self.setup_voltage()
        return self.VoltageIterator(self.sock, start, stop, step)

    def irange(self, start, stop, step) -> CurrentIterator:
        """Gets iterator to range of currents

        Parameters
        ----------
        start : float
            Starting current
        stop : float
            End current
        step : float
            Current step
        """

        self.setup_current()
        return self.CurrentIterator(self.sock, start, stop, step)

    def get_voltage(self) -> float:
        """Measure voltage supplied to the SPS from SMU

        Returns
        -------
        float
            Measured voltage
        """

        self.sock.sendall(b":FORM:ELEM VOLT\n")
        self.sock.sendall(b":READ?\n")
        # receive response
        reply = self.sock.recv(256)
        # strip trailing \r\n characters
        reply = reply.strip()
        # convert to float and return
        return float(reply)

    def get_current(self) -> float:
        """Measure current supplied to the SPS from SMU

        Returns
        -------
        float
            Measured current
        """

        self.sock.sendall(
            b":FORM:ELEM CURR\n"
        )  # replace with serial.write with socket.write commands, std library aviable, lots of example code
        self.sock.sendall(b":READ?\n")
        # receive response
        reply = self.sock.recv(256)
        # strip trailing \r\n characters
        reply = reply.strip()
        # convert to float and return
        return float(reply)


class Recorder:
    def __init__(self, serialport: str, host: str, port: int, delay: int = 1):
        """Recorder constructor

        Initializes the connection to smu and sps.

        Args:
            serialport: Serial port to sps
            host: Hostname/ip of smu
            port: TCP port to smu
            delay: Delay between smu step and power measurement in seconds
        """

        self.sps = SoilPowerSensorController(serialport)
        self.smu = SMULANController(host, port)
        self.delay = delay

    def record_voltage(
        self, start: float, stop: float, step: float, samples: int
    ) -> dict[str, list]:
        """Records voltage measurements from the smu and sps

        Input arguments given in volts units. Returned dictionary has keys for the
        expected voltage (number from python), actual voltage (value measured by
        the smu), and meas voltage (value measured by the sps).

        Args:
            start: Start voltage
            stop: Stop voltage (inclusive)
            step: Step between voltage
            samples: Number of samples taken at each voltage

        Returns:
            Dictionary in the following pandas compatable format:
            {
                "expected": [],
                "actual": [],
                "meas": [],
            }
        """

        data = {
            "expected": [],
            "actual": [],
            "meas": [],
        }

        for value in tqdm(self.smu.vrange(start, stop, step)):
            time.sleep(self.delay)
            for _ in range(samples):
                # expected
                data["expected"].append(value)
                # smu
                data["actual"].append(self.smu.get_voltage())
                # sps
                sps_v, _ = self.sps.get_power()
                data["meas"].append(sps_v)

        return data

    def record_current(
        self, start: float, stop: float, step: float, samples: int
    ) -> dict[str, list]:
        """Records current measurements from the smu and sps

        Input arguments given in amps units. Returned dictionary has keys for the
        expected current (number from python), actual current (value measured by
        the smu), and meas current (value measured by the sps).

        Args:
            start: Start current
            stop: Stop current (inclusive)
            step: Step between current
            samples: Number of samples taken at each voltage

        Returns:
            Dictionary in the following pandas compatable format:
            {
                "expected": [],
                "actual": [],
                "meas": [],
            }
        """

        data = {
            "expected": [],
            "actual": [],
            "meas": [],
        }

        for value in tqdm(self.smu.irange(start, stop, step)):
            time.sleep(self.delay)
            for _ in range(samples):
                # expected
                data["expected"].append(value)
                # smu
                data["actual"].append(self.smu.get_current())
                # sps
                _, sps_i = self.sps.get_power()
                data["meas"].append(sps_i)

        return data



"""To test connection via socket to Kiethley 2450

Stephen Taylor, 5/20/2024
"""

import socket


def ping_smu(host, port):
    try:
        # Create a TCP/IP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # Set a timeout for the connection attempt
        sock.settimeout(1)

        # Connect to the SMU
        sock.connect((host, port))

        # Send the *IDN? command
        sock.sendall(b"*RST\n")
        sock.sendall(b"*IDN?\n")

        # Receive the response
        response = sock.recv(1024)
        response = response.strip()
        print("Response from SMU:", response)

        # Close the socket
        sock.close()

        # Return True if connection successful
        return True
    except Exception as e:
        # Connection failed
        print(f"Error: {e}")
        return False


if __name__ == "__main__":
    # Specify the IP address or hostname of the SMU and its port
    smu_host = (
        "128.114.204.56"  # Replace with the actual IP address or hostname of the SMU
    )
    smu_port = 23  # Replace with the actual port used by the SMU

    # Ping the SMU
    if ping_smu(smu_host, smu_port):
        print("SMU is reachable.")
    else:
        print("Failed to ping SMU.")



import argparse

import os
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
import time

from .calibrate.recorder import Recorder
from .calibrate.linear_regression import (
    linear_regression,
    print_eval,
    print_coef,
    print_norm,
)
from .calibrate.plots import (
    plot_measurements,
    plot_calib,
    plot_residuals,
    plot_residuals_hist,
)

from .proto.encode import (
    encode_power_measurement,
    encode_phytos31_measurement,
    encode_teros12_measurement,
    encode_response,
)
from .proto.decode import decode_measurement, decode_response
from .proto.esp32 import encode_esp32command, decode_esp32command

from .proto.sensor import (
    encode_repeated_sensor_measurements,
    decode_repeated_sensor_measurements,
    encode_sensor_response,
    decode_sensor_response,
)

from .simulator.node import NodeSimulator


def entry():
    """Command line interface entry point"""

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(help="Ents Utilities", required=True)

    create_encode_parser(subparsers)
    create_encode_generic_parser(subparsers)
    create_decode_parser(subparsers)
    create_decode_generic_parser(subparsers)
    create_calib_parser(subparsers)
    create_sim_parser(subparsers)
    create_sim_generic_parser(subparsers)

    args = parser.parse_args()
    args.func(args)


def create_sim_generic_parser(subparsers):
    """Creates the generic simulation subparser

    Args:
        subparsers: Reference to subparser group
    Returns:
        Reference to new subparser
    """

    sim_p = subparsers.add_parser("sim_generic", help="Simluate generic sensor uploads")
    sim_p.add_argument(
        "--url",
        required=True,
        type=str,
        help="URL of the dirtviz instance (default: http://localhost:8000)",
    )
    sim_p.add_argument(
        "--mode",
        required=True,
        choices=["batch", "stream"],
        type=str,
        help="Upload mode",
    )
    sim_p.add_argument(
        "--sensor",
        required=True,
        type=str,
        nargs="+",
        help="Type of sensor to simulate",
    )
    sim_p.add_argument(
        "--min", type=float, default=-1.0, help="Minimum sensor value (default: -1.0)"
    )
    sim_p.add_argument(
        "--max", type=float, default=1.0, help="Maximum sensor value (default: 1.0)"
    )
    sim_p.add_argument("--cell", required=True, type=int, help="Cell Id")
    sim_p.add_argument("--logger", required=True, type=int, help="Logger Id")
    sim_p.add_argument("--start", type=str, help="Start date")
    sim_p.add_argument("--end", type=str, help="End date")
    sim_p.add_argument(
        "--freq", default=10.0, type=float, help="Frequency of uploads (default: 10s)"
    )
    sim_p.set_defaults(func=simulate_generic)

    return sim_p


def create_sim_parser(subparsers):
    """Creates the simulation subparser

    Args:
        subparsers: Reference to subparser group
    Returns:
        Reference to new subparser
    """

    sim_p = subparsers.add_parser("sim", help="Simluate sensor uploads")
    sim_p.add_argument(
        "--url",
        required=True,
        type=str,
        help="URL of the dirtviz instance (default: http://localhost:8000)",
    )
    sim_p.add_argument(
        "--mode",
        required=True,
        choices=["batch", "stream"],
        type=str,
        help="Upload mode",
    )
    sim_p.add_argument(
        "--sensor",
        required=True,
        choices=["power", "teros12", "teros21", "bme280"],
        type=str,
        nargs="+",
    )
    sim_p.add_argument("--cell", required=True, type=int, help="Cell Id")
    sim_p.add_argument("--logger", required=True, type=int, help="Logger Id")
    sim_p.add_argument("--start", type=str, help="Start date")
    sim_p.add_argument("--end", type=str, help="End date")
    sim_p.add_argument(
        "--freq", default=10.0, type=float, help="Frequency of uploads (default: 10s)"
    )
    sim_p.set_defaults(func=simulate)

    return sim_p


def simulate_generic(args):
    simulation = NodeSimulator(
        cell=args.cell,
        logger=args.logger,
        sensors=args.sensor,
        _min=args.min,
        _max=args.max,
    )

    if args.mode == "batch":
        if (args.start is None) or (args.end is None):
            raise ValueError("Start and end date must be provided for batch mode.")

        # format dates
        curr_dt = datetime.fromisoformat(args.start)
        end_dt = datetime.fromisoformat(args.end)

        # create list of measurements
        while curr_dt <= end_dt:
            ts = int(curr_dt.timestamp())
            simulation.measure(ts)
            curr_dt += timedelta(seconds=args.freq)

        # send measurements
        while simulation.send_next(args.url):
            print(simulation)

        print("Done!")

    elif args.mode == "stream":
        print("Use CTRL+C to stop the simulation")
        try:
            while True:
                dt = datetime.now()
                ts = int(dt.timestamp())
                simulation.measure(ts)
                while simulation.send_next(args.url):
                    print(simulation)
                time.sleep(args.freq)
        except KeyboardInterrupt as _:
            print("Stopping simulation")


def simulate(args):
    simulation = NodeSimulator(
        cell=args.cell,
        logger=args.logger,
        sensors=args.sensor,
    )

    if args.mode == "batch":
        if (args.start is None) or (args.end is None):
            raise ValueError("Start and end date must be provided for batch mode.")

        # format dates
        curr_dt = datetime.fromisoformat(args.start)
        end_dt = datetime.fromisoformat(args.end)

        # create list of measurements
        while curr_dt <= end_dt:
            ts = int(curr_dt.timestamp())
            simulation.measure(ts)
            curr_dt += timedelta(seconds=args.freq)

        # send measurements
        while simulation.send_next(args.url):
            print(simulation)

        print("Done!")

    elif args.mode == "stream":
        print("Use CTRL+C to stop the simulation")
        try:
            while True:
                dt = datetime.now()
                ts = int(dt.timestamp())
                simulation.measure(ts)
                while simulation.send_next(args.url):
                    print(simulation)
                time.sleep(args.freq)
        except KeyboardInterrupt as _:
            print("Stopping simulation")


def create_calib_parser(subparsers):
    """Creates the calibration subparser

    Args:
        subparsers: Reference to subparser group
    Returns:
        Reference to new subparser
    """

    # calibration parser
    calib_p = subparsers.add_parser("calib", help="Calibrate power measurements")
    calib_p.add_argument(
        "--samples",
        type=int,
        default=10,
        required=False,
        help="Samples taken at each step (default: 10)",
    )
    calib_p.add_argument(
        "--plot", action="store_true", help="Show calibration parameter plots"
    )
    calib_p.add_argument(
        "--mode",
        type=str,
        default="both",
        required=False,
        help="Either both, voltage, or current (default: both)",
    )
    calib_p.add_argument(
        "--output", type=str, required=False, help="Output directory for measurements"
    )
    calib_p.add_argument("port", type=str, help="Board serial port")
    calib_p.add_argument("host", type=str, help="Address and port of smu (ip:port)")
    calib_p.set_defaults(func=calibrate)

    return calib_p


def create_encode_generic_parser(subparsers):
    """Create generic encode command subparser

    Args:
        subparsers: Reference to subparser group

    Returns:
        Reference to new subparser
    """

    encode_parser = subparsers.add_parser("encode_generic", help="Encode generic data")

    print_format = encode_parser.add_mutually_exclusive_group()
    print_format.add_argument(
        "--hex", action="store_true", help="Print as hex values (default)"
    )
    print_format.add_argument(
        "--raw", action="store_true", help="Print raw bytes object"
    )
    print_format.add_argument(
        "--c", action="store_true", help="Print bytes for copying to c"
    )

    encode_subparsers = encode_parser.add_subparsers(
        title="Message type", dest="type", required=True
    )

    # sensor measurement
    measurement_parser = encode_subparsers.add_parser(
        "meas", help='Proto "Measurement" message'
    )
    measurement_parser.add_argument("--ts", type=int, help="Unix epoch timestamp")
    measurement_parser.add_argument("--cell", type=int, help="Cell Id")
    measurement_parser.add_argument("--logger", type=int, help="Logger Id")
    measurement_parser.add_argument("--idx", type=int, default=1, help="Upload index")
    measurement_parser.add_argument(
        "--sensor",
        nargs=2,
        metavar=("type", "value"),
        action="append",
        required=True,
        help="Specify as: --sensor <type> <value>",
    )
    measurement_parser.set_defaults(func=handle_encode_generic_measurement)

    # response
    response_parser = encode_subparsers.add_parser(
        "resp", help='Proto "Response" message'
    )
    response_parser.add_argument(
        "--resp",
        nargs=2,
        metavar=("idx", "status"),
        action="append",
        required=True,
        help="Specify as: --resp <idx> <status>",
    )
    response_parser.set_defaults(func=handle_encode_generic_response)

    return encode_parser


def parse_number(s: str) -> tuple:
    try:
        i = int(s)
        if i >= 0:
            return i, "unsignedInt"
        return i, "signedInt"
    except ValueError:
        pass

    try:
        return float(s), "decimal"
    except ValueError:
        pass

    raise ValueError(f"Invalid numeric value: {s}")


def handle_encode_generic_measurement(args):
    """Take arguments and encode a repeated senosr measurement message"""

    meas = {
        "meta": {
            "ts": args.ts,
            "cellId": args.cell,
            "loggerId": args.logger,
        },
        "measurements": [],
    }

    for s in args.sensor:
        val, val_type = parse_number(s[1])
        meas["measurements"].append(
            {
                "type": s[0],
                val_type: val,
                "idx": args.idx,
            }
        )
        args.idx += 1

    data = encode_repeated_sensor_measurements(meas)
    print_data(args, data)


def handle_encode_generic_response(args):
    """Takes arguments and encode a sensor response message"""

    resp = {
        "responses": [],
    }

    for r in args.resp:
        idx = int(r[0])
        status = r[1]
        resp["responses"].append(
            {
                "uploadIndex": idx,
                "status": status,
            }
        )

    data = encode_sensor_response(resp)
    print_data(args, data)


def create_encode_parser(subparsers):
    """Create encode command subparser

    Args:
        subparsers: Reference to subparser group

    Returns:
        Reference to new subparser
    """

    encode_parser = subparsers.add_parser("encode", help="Encode data")

    print_format = encode_parser.add_mutually_exclusive_group()
    print_format.add_argument(
        "--hex", action="store_true", help="Print as hex values (default)"
    )
    print_format.add_argument(
        "--raw", action="store_true", help="Print raw bytes object"
    )
    print_format.add_argument(
        "--c", action="store_true", help="Print bytes for copying to c"
    )

    encode_subparsers = encode_parser.add_subparsers(
        title="Message type",
        dest="type",
        required=True,
    )

    def create_measurement_parser(encode_subparsers):
        measurement_parser = encode_subparsers.add_parser(
            "measurement", help='Proto "Measurement" message'
        )
        measurement_subparser = measurement_parser.add_subparsers(
            title="Measurement type",
            required=True,
        )

        # metadata
        measurement_parser.add_argument("ts", type=int, help="Unix epoch timestamp")
        measurement_parser.add_argument("cell", type=int, help="Cell Id")
        measurement_parser.add_argument("logger", type=int, help="Logger Id")

        power_parser = measurement_subparser.add_parser(
            "power", help="PowerMeasurement"
        )
        power_parser.add_argument("voltage", type=float, help="Voltage in (V)")
        power_parser.add_argument("current", type=float, help="Current in (A)")
        power_parser.set_defaults(func=handle_encode_measurement_power)

        teros12_parser = measurement_subparser.add_parser(
            "teros12", help="Teros12Measurement"
        )
        teros12_parser.add_argument("vwc_raw", type=float, help="Raw vwc")
        teros12_parser.add_argument("vwc_adj", type=float, help="Calibrated vwc")
        teros12_parser.add_argument("temp", type=float, help="Temperature in C")
        teros12_parser.add_argument("ec", type=int, help="Electrical conductivity")
        teros12_parser.set_defaults(func=handle_encode_measurement_teros12)

        phytos31_parser = measurement_subparser.add_parser(
            "phytos31", help="Phytos31Measurement"
        )
        phytos31_parser.add_argument("voltage", type=float, help="Raw voltage (V)")
        phytos31_parser.add_argument("leaf_wetness", type=float, help="Leaf wetness")
        phytos31_parser.set_defaults(func=handle_encode_measurement_phytos31)

        return measurement_parser

    def create_response_parser(encode_subparsers):
        response_parser = encode_subparsers.add_parser(
            "response", help='Proto "Response" message'
        )
        response_parser.add_argument("status", type=str, help="Status")
        response_parser.set_defaults(func=handle_encode_response)

        return response_parser

    def create_esp32command_parser(encode_subparsers):
        esp32command_parser = encode_subparsers.add_parser(
            "esp32command", help='Proto "Esp32Command" message'
        )
        esp32command_subparser = esp32command_parser.add_subparsers(
            title="type",
            help="PageCommand",
            required=True,
        )

        test_parser = esp32command_subparser.add_parser("test", help="TestCommand")
        test_parser.add_argument("state", type=str, help="State to put module into")
        test_parser.add_argument("data", type=int, help="Data associated with command")
        test_parser.set_defaults(func=handle_encode_esp32command_test)

        esp32_parser = esp32command_subparser.add_parser("page", help="PageCommand")
        esp32_parser.add_argument("type", type=str, help="Request type")
        esp32_parser.add_argument("fd", type=int, help="File descriptor")
        esp32_parser.add_argument("bs", type=int, help="Block size")
        esp32_parser.add_argument("num", type=int, help="Number of bytes")
        esp32_parser.set_defaults(func=handle_encode_esp32command_page)

        wifi_parser = esp32command_subparser.add_parser("wifi", help="WiFiCommand")
        wifi_parser.add_argument("type", type=str, help="WiFi command type")
        wifi_parser.add_argument("--ssid", type=str, default="", help="WiFi SSID")
        wifi_parser.add_argument("--passwd", type=str, default="", help="WiFi password")
        wifi_parser.add_argument("--url", type=str, default="", help="Endpoint url")
        wifi_parser.add_argument("--port", type=int, default=0, help="Endpoint port")
        wifi_parser.add_argument("--rc", type=int, default=0, help="Return code")
        wifi_parser.add_argument("--ts", type=int, help="Timestamp in unix epochs")
        wifi_parser.add_argument(
            "--resp", type=str, default=b"", help="Serialized response message"
        )
        wifi_parser.set_defaults(func=handle_encode_esp32command_wifi)

        return esp32command_parser

    # create subparsers
    create_measurement_parser(encode_subparsers)
    create_response_parser(encode_subparsers)
    create_esp32command_parser(encode_subparsers)

    return encode_parser


def handle_encode_measurement_power(args):
    data = encode_power_measurement(
        ts=args.ts,
        cell_id=args.cell,
        logger_id=args.logger,
        voltage=args.voltage,
        current=args.current,
    )

    print_data(args, data)


def handle_encode_measurement_teros12(args):
    data = encode_teros12_measurement(
        ts=args.ts,
        cell_id=args.cell,
        logger_id=args.logger,
        vwc_raw=args.vwc_raw,
        vwc_adj=args.vwc_adj,
        temp=args.temp,
        ec=args.ec,
    )

    print_data(args, data)


def handle_encode_measurement_phytos31(args):
    data = encode_phytos31_measurement(
        ts=args.ts,
        cell_id=args.cell,
        logger_id=args.logger,
        voltage=args.voltage,
        leaf_wetness=args.leaf_wetness,
    )

    print_data(args, data)


def handle_encode_response(args):
    valid_status = ["SUCCESS", "ERROR"]
    if args.status not in valid_status:
        raise NotImplementedError(f'Response status "{args.status}" not implemented')

    data = encode_response(args.status)
    print_data(args, data)


def handle_encode_esp32command_test(args):
    data = encode_esp32command("test", state=args.state, data=args.data)
    print_data(args, data)


def handle_encode_esp32command_page(args):
    data = encode_esp32command(
        "page", req=args.type.lower(), fd=args.fd, bs=args.bs, n=args.num
    )
    print_data(args, data)


def handle_encode_esp32command_wifi(args):
    data = encode_esp32command(
        "wifi",
        _type=args.type.lower(),
        ssid=args.ssid,
        passwd=args.passwd,
        url=args.url,
        port=args.port,
        rc=args.rc,
        ts=args.ts,
        resp=args.resp,
    )
    print_data(args, data)


def print_data(args, data: bytes) -> str:
    if args.c:
        print_bytes_c(data)
    elif args.raw:
        print(data)
    else:
        print(data.hex())


def print_bytes_c(data: bytes) -> str:
    """Formats serialized data into c bytes array"""

    # format data string
    data_str = "uint8_t data[] = {"
    hex_str = [hex(d) for d in data]
    data_str += ", ".join(hex_str)
    data_str += "};"

    # print data string
    print(data_str)

    # print length of data string
    print(f"size_t data_len = {len(hex_str)};")


def create_decode_generic_parser(subparsers):
    """Create generic decode command parser

    Args:
        subparsers: Reference to subparser group

    Returns:
        Reference to new subparser
    """

    decode_parser = subparsers.add_parser("decode_generic", help="Decode generic data")

    decode_subparsers = decode_parser.add_subparsers(
        title="Message type",
        dest="type",
        required=True,
    )

    decode_parser.add_argument(
        "data", type=str, help="Protobuf serialized data in hex format"
    )

    # sensor measurement
    measurement_parser = decode_subparsers.add_parser(
        "meas", help='Proto "Measurement" message'
    )
    measurement_parser.set_defaults(func=handle_decode_generic_measurement)

    # response
    response_parser = decode_subparsers.add_parser(
        "resp", help='Proto "Response" message'
    )
    response_parser.set_defaults(func=handle_decode_generic_response)

    return decode_parser


def handle_decode_generic_measurement(args):
    data = bytes.fromhex(args.data)
    vals = decode_repeated_sensor_measurements(data)
    print(vals)


def handle_decode_generic_response(args):
    data = bytes.fromhex(args.data)
    vals = decode_sensor_response(data)
    print(vals)


def create_decode_parser(subparsers):
    """Create decode command parser

    Args:
        subparsers: Reference to subparser group

    Returns:
        Reference to new subparser
    """

    decode_parser = subparsers.add_parser("decode", help="Decode data")

    decode_subparsers = decode_parser.add_subparsers(
        title="Message type",
        dest="type",
        required=True,
    )

    # measurement
    measurement_parser = decode_subparsers.add_parser(
        "measurement", help='Proto "Measurement" message'
    )
    measurement_parser.set_defaults(func=handle_decode_measurement)

    # response
    response_parser = decode_subparsers.add_parser(
        "response", help='Proto "Response" message'
    )
    response_parser.set_defaults(func=handle_decode_response)

    # esp32command
    esp32command_parser = decode_subparsers.add_parser(
        "esp32command", help='Proto "Esp32Command" message'
    )
    esp32command_parser.set_defaults(func=handle_decode_esp32command)

    decode_parser.add_argument(
        "data", type=str, help="Protobuf serialized data in hex format"
    )

    return decode_parser


def handle_decode_measurement(args):
    data = bytes.fromhex(args.data)
    vals = decode_measurement(data)
    print(vals)


def handle_decode_response(args):
    data = bytes.fromhex(args.data)
    vals = decode_response(data)
    print(vals)


def handle_decode_esp32command(args):
    data = bytes.fromhex(args.data)
    vals = decode_esp32command(data)
    print(vals)


def calibrate(args):
    print(
        "If you don't see any output for 5 seconds, restart the calibration after resetting the ents board"
    )

    host, port = args.host.split(":")
    rec = Recorder(args.port, host, int(port))

    if args.mode == "both":
        run_v = True
        run_i = True
    elif args.mode in ["v", "volts", "voltage"]:
        run_v = True
        run_i = False
    elif args.mode in ["i", "amps", "current"]:
        run_v = True
        run_i = True
    else:
        raise NotImplementedError(f"Calbration mode: {args.mode} not implemented")

    V_START = -2.0
    V_STOP = 2.0
    V_STEP = 0.5

    I_START = -0.0009
    I_STOP = 0.0009
    I_STEP = 0.00045

    def record_calibrate(start, stop, step, name: str):
        """Record and calibrate

        Args:
            start: Start value
            stop: Stop value (inclusive)
            step: Step between values
            name: Name of channel
        """

        # TODO Unjank reference to member variables by moving the selection to
        # the class.
        if name == "voltage":
            iterator = Recorder.record_voltage
        elif name == "current":
            iterator = Recorder.record_current

        # collect data
        print("Collecting calibration data")
        cal = iterator(rec, start, stop, step, args.samples)
        if args.output:
            save_csv(cal, args.output, f"{name}-cal.csv")

        print("Collecting evaluation data")
        _eval = iterator(rec, start, stop, step, args.samples)
        if args.output:
            save_csv(_eval, args.output, f"{name}-eval.csv")

        model = linear_regression(
            np.array(cal["meas"]).reshape(-1, 1), np.array(cal["actual"]).reshape(-1, 1)
        )
        pred = model.predict(np.array(_eval["meas"]).reshape(-1, 1))
        residuals = np.array(_eval["actual"]) - pred.flatten()

        print("")
        print("\r\rnCoefficients")
        print_coef(model)
        print("\r\nEvaluation")
        print_eval(pred, _eval["actual"])
        print("\r\nNormal fit")
        print_norm(residuals)
        print("")

        # plots
        if args.plot:
            plot_measurements(cal["actual"], cal["meas"], title=name)
            plot_calib(_eval["meas"], pred, title=name)
            plot_residuals(pred, residuals, title=name)
            plot_residuals_hist(residuals, title=name)

    if run_v:
        print("Connect smu to voltage inputs device and press ENTER")
        input()
        record_calibrate(V_START, V_STOP, V_STEP, "voltage")

    if run_i:
        print(
            "Connect smu to a resistor in series with the current channels and press ENTER"
        )
        input()
        record_calibrate(I_START, I_STOP, I_STEP, "current")

    print("Press enter to close plots")
    input()


def save_csv(data: dict[str, list], path: str, name: str):
    """Save measurement dictionary to csv

    Args:
        data: Measurement data
        path: Folder path
        name: Name of csv file
    """
    path = os.path.join(path, name)
    pd.DataFrame(data).to_csv(path, index=False)



"""Tests encoding and decoding of protobuf serialized data

Encoding for Response messages checks both a SUCCESS and ERROR can be obtained.
Decoding is performed to ensure data is preserved.

Decoding checks that the Measurement message is preserved through an
encoding/decoding cycle. Checks that missing fields result in an error and the
correct dictionary format is returned.
"""

import unittest
import base64

from ents.proto import (
    encode_response,
    decode_measurement,
    encode_esp32command,
    decode_esp32command,
)

from ents.proto.soil_power_sensor_pb2 import (
    Measurement,
    Response,
    MeasurementMetadata,
    Esp32Command,
    PageCommand,
    TestCommand,
)


class TestEncode(unittest.TestCase):
    def test_success(self):
        # encode
        resp_str = encode_response(success=True)

        # decode
        resp_out = Response()
        resp_out.ParseFromString(resp_str)

        self.assertEqual(Response.ResponseType.SUCCESS, resp_out.resp)

    def test_error(self):
        # encode
        resp_str = encode_response(success=False)

        # decode
        resp_out = Response()
        resp_out.ParseFromString(resp_str)

        self.assertEqual(Response.ResponseType.ERROR, resp_out.resp)


class TestDecode(unittest.TestCase):
    """Test decoding of measurements"""

    def setUp(self):
        """Creates a default metadata message"""
        self.meta = MeasurementMetadata()
        self.meta.ts = 1436079600
        self.meta.cell_id = 20
        self.meta.logger_id = 4

    def check_meta(self, meas_dict: dict):
        """Checks the measurement dictionary contains metadata information"""

        self.assertEqual(1436079600, meas_dict["ts"])
        self.assertEqual(20, meas_dict["cellId"])
        self.assertEqual(4, meas_dict["loggerId"])

    def test_power(self):
        """Test decoding of PowerMeasurement"""

        # import pdb; pdb.set_trace()
        # format measurement
        meas = Measurement()
        meas.meta.CopyFrom(self.meta)
        meas.power.voltage = 122.38
        meas.power.current = 514.81

        # serialize
        meas_str = meas.SerializeToString()

        # decode
        meas_dict = decode_measurement(data=meas_str)

        # check resulting dict
        self.assertEqual("power", meas_dict["type"])
        self.check_meta(meas_dict)
        self.assertAlmostEqual(122.38, meas_dict["data"]["voltage"])
        self.assertEqual(float, meas_dict["data_type"]["voltage"])
        self.assertAlmostEqual(514.81, meas_dict["data"]["current"])
        self.assertEqual(float, meas_dict["data_type"]["current"])

    def test_teros12(self):
        """Test decoding of Teros12Measurement"""

        # format measurement
        meas = Measurement()
        meas.meta.CopyFrom(self.meta)
        meas.teros12.vwc_raw = 2124.62
        meas.teros12.vwc_adj = 0.43
        meas.teros12.temp = 24.8
        meas.teros12.ec = 123

        # serialize
        meas_str = meas.SerializeToString()

        # decode
        meas_dict = decode_measurement(data=meas_str)

        # check dict
        self.assertEqual("teros12", meas_dict["type"])
        self.check_meta(meas_dict)
        self.assertAlmostEqual(2124.62, meas_dict["data"]["vwcRaw"])
        self.assertEqual(float, meas_dict["data_type"]["vwcRaw"])
        self.assertAlmostEqual(0.43, meas_dict["data"]["vwcAdj"])
        self.assertEqual(float, meas_dict["data_type"]["vwcAdj"])
        self.assertAlmostEqual(24.8, meas_dict["data"]["temp"])
        self.assertEqual(float, meas_dict["data_type"]["temp"])
        self.assertEqual(123, meas_dict["data"]["ec"])
        self.assertEqual(int, meas_dict["data_type"]["ec"])

    def test_bme280(self):
        """Test decoding of bme280 measurement"""

        meas = Measurement()
        meas.meta.CopyFrom(self.meta)
        meas.bme280.pressure = 98473
        meas.bme280.temperature = 2275
        meas.bme280.humidity = 43600

        # serialize
        meas_str = meas.SerializeToString()

        # decode
        meas_dict = decode_measurement(data=meas_str)

        # check dict
        self.assertEqual("bme280", meas_dict["type"])
        self.check_meta(meas_dict)
        self.assertEqual(98473, meas_dict["data"]["pressure"])
        self.assertEqual(int, meas_dict["data_type"]["pressure"])
        self.assertEqual(2275, meas_dict["data"]["temperature"])
        self.assertEqual(int, meas_dict["data_type"]["temperature"])
        self.assertEqual(43600, meas_dict["data"]["humidity"])
        self.assertEqual(int, meas_dict["data_type"]["humidity"])

        # decode
        meas_dict = decode_measurement(data=meas_str, raw=False)

        # check dict
        self.assertEqual("bme280", meas_dict["type"])
        self.check_meta(meas_dict)
        self.assertAlmostEqual(9847.3, meas_dict["data"]["pressure"])
        self.assertEqual(float, meas_dict["data_type"]["pressure"])
        self.assertAlmostEqual(22.75, meas_dict["data"]["temperature"])
        self.assertEqual(float, meas_dict["data_type"]["temperature"])
        self.assertAlmostEqual(43.600, meas_dict["data"]["humidity"])
        self.assertEqual(float, meas_dict["data_type"]["humidity"])

    def test_teros21(self):
        meas = Measurement()
        meas.meta.CopyFrom(self.meta)

        meas.teros21.matric_pot = 101.23
        meas.teros21.temp = 22.50

        # serialize
        meas_str = meas.SerializeToString()

        # decode
        meas_dict = decode_measurement(data=meas_str)

        # check dict
        self.assertEqual("teros21", meas_dict["type"])
        self.check_meta(meas_dict)
        self.assertAlmostEqual(101.23, meas_dict["data"]["matricPot"])
        self.assertEqual(float, meas_dict["data_type"]["matricPot"])
        self.assertAlmostEqual(22.50, meas_dict["data"]["temp"])
        self.assertEqual(float, meas_dict["data_type"]["temp"])

    def test_missing_meta(self):
        """Test that error is raised when meta is not set"""

        # format measurement
        meas = Measurement()
        meas.power.voltage = 122.38
        meas.power.current = 514.81

        # serialize
        meas_str = meas.SerializeToString()

        # decode
        with self.assertRaises(KeyError):
            decode_measurement(data=meas_str)

    def test_missing_measurement(self):
        """Test that error is raised when measurement is missing"""

        # format measurement
        meas = Measurement()
        meas.meta.CopyFrom(self.meta)

        # serialize
        meas_str = meas.SerializeToString()

        # decode
        with self.assertRaises(KeyError):
            decode_measurement(data=meas_str)

    def test_missing_default_values(self):
        """Test to ensure that default valued fields are included in the
        decoded dictionary
        """

        meas = Measurement()
        meas.meta.CopyFrom(self.meta)

        meas.teros12.vwc_raw = 2141.52
        meas.teros12.vwc_adj = 0.45
        meas.teros12.temp = 25.0

        # set integer to default value of 0
        meas.teros12.ec = 0

        meas_str = meas.SerializeToString()

        meas_dict = decode_measurement(data=meas_str)

        self.assertIn("ec", meas_dict["data"])


class TestEsp32(unittest.TestCase):
    def test_cmd_not_implemented(self):
        """Checks that an exception is raised when a non-existing command is
        called"""

        with self.assertRaises(NotImplementedError):
            encode_esp32command("agg", req="open", fd=123, bs=456, n=789)

    def test_page_encode(self):
        """Test encoding a page command"""

        req = "open"
        fd = 1
        bs = 512
        n = 1024

        cmd_str = encode_esp32command("page", req=req, fd=fd, bs=bs, n=n)

        cmd = Esp32Command()
        cmd.ParseFromString(cmd_str)

        # check the command type
        cmd_type = cmd.WhichOneof("command")
        self.assertEqual(cmd_type, "page_command")

        # check individual values
        self.assertEqual(cmd.page_command.file_request, PageCommand.RequestType.OPEN)
        self.assertEqual(cmd.page_command.file_descriptor, fd)
        self.assertEqual(cmd.page_command.block_size, bs)
        self.assertEqual(cmd.page_command.num_bytes, n)

    def test_page_decode(self):
        """Test decoding a page command"""

        req = "open"
        fd = 1
        bs = 512
        n = 1024

        cmd_str = encode_esp32command("page", req=req, fd=fd, bs=bs, n=n)

        cmd = decode_esp32command(cmd_str)

        # check page command
        self.assertIn("pageCommand", cmd)

        cmd = cmd["pageCommand"]

        # check individual values
        self.assertEqual(cmd["fileRequest"], "OPEN")
        self.assertEqual(cmd["fileDescriptor"], fd)
        self.assertEqual(cmd["blockSize"], bs)
        self.assertEqual(cmd["numBytes"], n)

    def test_page_req_not_implemented(self):
        """Test encoding a page command with a not implemented request"""

        with self.assertRaises(NotImplementedError):
            encode_esp32command("page", req="agg", fd=123, bs=456, n=789)

    def test_test_encode(self):
        """Test encoding a test command"""

        state = "receive"
        num = 123

        cmd_str = encode_esp32command("test", state=state, data=num)

        cmd = Esp32Command()
        cmd.ParseFromString(cmd_str)

        # check the command type
        cmd_type = cmd.WhichOneof("command")
        self.assertEqual(cmd_type, "test_command")

        # check individual values
        self.assertEqual(cmd.test_command.state, TestCommand.ChangeState.RECEIVE)
        self.assertEqual(cmd.test_command.data, num)

    def test_test_decode(self):
        """Test decoding a test command"""

        state = "receive"
        num = 123

        cmd_str = encode_esp32command("test", state=state, data=num)

        cmd = decode_esp32command(cmd_str)

        # check test command
        self.assertIn("testCommand", cmd)

        cmd = cmd["testCommand"]

        # check individual values
        self.assertEqual(cmd["state"], "RECEIVE")
        self.assertEqual(cmd["data"], num)

    def test_test_state_not_implemented(self):
        """Test encoding a page command with a not implemented state"""

        with self.assertRaises(NotImplementedError):
            encode_esp32command("test", state="agg", data=123)

    def test_wifi(self):
        """Test encoding/decoding a WiFi command

        All parameters are tested at once. This is not the intended
        implementation.
        """

        _type = "POST"
        ssid = "HelloWorld"
        passwd = "password"
        url = "https://test.com"
        port = 6969
        rc = 200
        ts = 1652346246
        resp = b"agga"

        # encode
        cmd_str = encode_esp32command(
            "wifi",
            _type=_type,
            ssid=ssid,
            passwd=passwd,
            url=url,
            port=port,
            rc=rc,
            ts=ts,
            resp=resp,
        )

        # decode
        cmd = decode_esp32command(cmd_str)

        # check the command key exists
        self.assertIn("wifiCommand", cmd)

        cmd = cmd["wifiCommand"]

        # check individual values
        self.assertEqual(cmd["type"], _type)
        self.assertEqual(cmd["ssid"], ssid)
        self.assertEqual(cmd["passwd"], passwd)
        self.assertEqual(cmd["url"], url)
        self.assertEqual(cmd["port"], port)
        self.assertEqual(cmd["rc"], rc)
        self.assertEqual(cmd["ts"], ts)
        self.assertEqual(base64.b64decode(cmd["resp"]), resp)

    def test_wifi_type_not_implemented(self):
        """Encode a WiFiCommand with a improper type to check for
        NotImplementedError
        """

        with self.assertRaises(NotImplementedError):
            encode_esp32command("wifi", _type="bla")


if __name__ == "__main__":
    unittest.main()



import numpy as np
from sklearn import linear_model
from sklearn.metrics import (
    mean_absolute_error,
    mean_squared_error,
    r2_score,
    mean_absolute_percentage_error,
)
from scipy.stats import norm


def linear_regression(meas: list, actual: list):
    """Performs linear regression between measured and actual values

    Args:
        meas: Measure values
        actual: Ground truth values

    Returns:
        Linear regression model
    """

    model = linear_model.LinearRegression()
    model.fit(meas, actual)

    return model


def print_eval(pred: list, actual: list):
    """Prints various evaluation parameters of the model

    The following metrics are used:
        - Mean absolute error
        - Root mean square error
        - R2 score
        - Mean absolute percentage error

    Args:
        pred: Predicted values
        actual: Actual values
    """
    mae = mean_absolute_error(actual, pred)
    print(f"Mean absolute error: {mae:.4f}")

    rmse = np.sqrt(mean_squared_error(actual, pred))
    print(f"Root mean square error: {rmse:.4f}")

    r2 = r2_score(actual, pred)
    print(f"R-squared: {r2:.4f}")

    mape = mean_absolute_percentage_error(actual, pred)
    print(f"Mean absolute percentage error: {mape:.4f}")


def print_coef(model):
    """Print coefficients of a model

    Args:
        model: Linear model
    """

    slope = model.coef_
    inter = model.intercept_

    print(f"Slope: {slope}")
    print(f"Intercept: {inter}")


def print_norm(residuals: list):
    """Print mean offset and standard deviation

    Args:
        residuals: List of residuals
    """

    mu, std = norm.fit(residuals)
    print(f"mu = {mu}")
    print(f"std = {std}")



"""Tests encoding/decoding of sensor data."""

import unittest

from ents.proto.sensor import (
    encode_sensor_measurement,
    encode_repeated_sensor_measurements,
    decode_sensor_measurement,
    decode_repeated_sensor_measurements,
    update_repeated_metadata,
    get_sensor_data,
)


class TestProtoSensor(unittest.TestCase):
    """Tests encoding/decoding of sensor data."""

    def test_sensor_measurement(self):
        """Tests encoding/decoding of single SensorMeasurement."""

        # encode

        meas_in = {
            "meta": {
                "ts": 123,
                "loggerId": 456,
                "cellId": 789,
            },
            "type": "POWER_VOLTAGE",
            "unsignedInt": 100,
        }

        serialized = encode_sensor_measurement(meas_in)

        # decode
        meas_out = decode_sensor_measurement(serialized)

        self.assertEqual(meas_in, meas_out)

    def test_repeated_sensor_measurements(self):
        """Tests encoding/decoding of RepeatedSensorMeasurement."""

        meas_in = {
            "meta": {
                "ts": 123,
                "loggerId": 456,
                "cellId": 789,
            },
            "measurements": [
                {
                    "meta": {
                        "ts": 124,
                        "loggerId": 457,
                        "cellId": 790,
                    },
                    "type": "POWER_VOLTAGE",
                    "unsignedInt": 100,
                },
                {
                    "meta": {
                        "ts": 125,
                        "loggerId": 458,
                        "cellId": 791,
                    },
                    "type": "POWER_CURRENT",
                    "signedInt": -100,
                },
            ],
        }

        meas_expected = {
            "meta": {
                "ts": 123,
                "loggerId": 456,
                "cellId": 789,
            },
            "measurements": [
                {
                    "meta": {
                        "ts": 124,
                        "loggerId": 457,
                        "cellId": 790,
                    },
                    "type": "POWER_VOLTAGE",
                    "unsignedInt": 100,
                },
                {
                    "meta": {
                        "ts": 125,
                        "loggerId": 458,
                        "cellId": 791,
                    },
                    "type": "POWER_CURRENT",
                    "signedInt": -100,
                },
            ],
        }

        serialized = encode_repeated_sensor_measurements(meas_in)

        meas_out = decode_repeated_sensor_measurements(serialized)

        self.assertEqual(meas_expected, meas_out)

    def test_update_repeated_metadata_no_change(self):
        """Tests when no change should occur in the measurement."""

        meas_in = {
            "meta": {
                "ts": 123,
                "loggerId": 456,
                "cellId": 789,
            },
            "measurements": [
                {
                    "meta": {
                        "ts": 124,
                        "loggerId": 457,
                        "cellId": 790,
                    },
                    "type": "POWER_VOLTAGE",
                    "unsignedInt": 100,
                },
                {
                    "meta": {
                        "ts": 125,
                        "loggerId": 458,
                        "cellId": 791,
                    },
                    "type": "POWER_CURRENT",
                    "signedInt": -100,
                },
            ],
        }

        meas_out = update_repeated_metadata(meas_in)

        self.assertEqual(meas_in, meas_out)

    def test_update_repeated_metadata_ok(self):
        """Tests the typical case of updating repeated measurement metadata."""

        # top level metadata for comparison
        top_meta = {
            "ts": 123,
            "loggerId": 456,
            "cellId": 789,
        }

        meas_in = {
            "meta": {
                "ts": 123,
                "loggerId": 456,
                "cellId": 789,
            },
            "measurements": [
                {
                    "type": "POWER_VOLTAGE",
                    "unsignedInt": 100,
                },
                {
                    "meta": {
                        "ts": 125,
                        "loggerId": 458,
                        "cellId": 791,
                    },
                    "type": "POWER_CURRENT",
                    "signedInt": -100,
                },
            ],
        }

        meas_out = update_repeated_metadata(meas_in)

        self.assertIn("meta", meas_out["measurements"][0])
        self.assertEqual(meas_out["measurements"][0]["meta"], top_meta)

    def test_update_repeated_metadata_value_error(self):
        """Tests ValueError when single repeated measurement is missing
        metadata and there is no top level metadata to pull from."""

        meas_in = {
            "measurements": [
                {
                    "meta": {
                        "ts": 124,
                        "loggerId": 457,
                        "cellId": 790,
                    },
                    "type": "POWER_VOLTAGE",
                    "unsignedInt": 100,
                },
                {
                    "type": "POWER_CURRENT",
                    "signedInt": -100,
                },
            ]
        }

        with self.assertRaises(ValueError):
            _ = update_repeated_metadata(meas_in)

    def test_get_sensor_data(self):
        """Tests get_sensor_data function."""

        meas_type = "POWER_VOLTAGE"

        sensor_data = get_sensor_data(meas_type)

        expected_data = {
            "name": "Voltage",
            "unit": "mV",
        }

        self.assertEqual(sensor_data, expected_data)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: soil_power_sensor.proto
# Protobuf Python Version: 6.33.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    1,
    '',
    'soil_power_sensor.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17soil_power_sensor.proto\"E\n\x13MeasurementMetadata\x12\x0f\n\x07\x63\x65ll_id\x18\x01 \x01(\r\x12\x11\n\tlogger_id\x18\x02 \x01(\r\x12\n\n\x02ts\x18\x03 \x01(\r\"4\n\x10PowerMeasurement\x12\x0f\n\x07voltage\x18\x02 \x01(\x01\x12\x0f\n\x07\x63urrent\x18\x03 \x01(\x01\"*\n\x17VoltageDeltaMeasurement\x12\x0f\n\x07voltage\x18\x01 \x01(\r\"*\n\x17\x43urrentDeltaMeasurement\x12\x0f\n\x07\x63urrent\x18\x01 \x01(\r\"%\n\x12VoltageMeasurement\x12\x0f\n\x07voltage\x18\x01 \x01(\x01\"%\n\x12\x43urrentMeasurement\x12\x0f\n\x07\x63urrent\x18\x01 \x01(\x01\"K\n\x0fPowerDeltaEntry\x12\n\n\x02ts\x18\x01 \x01(\r\x12\x15\n\rvoltage_delta\x18\x02 \x01(\r\x12\x15\n\rcurrent_delta\x18\x03 \x01(\r\"E\n\x15PowerMeasurementDelta\x12\x15\n\rvoltage_delta\x18\x02 \x01(\r\x12\x15\n\rcurrent_delta\x18\x03 \x01(\r\"\\\n\x13RepeatedPowerDeltas\x12\x11\n\tlogger_id\x18\x01 \x01(\r\x12\x0f\n\x07\x63\x65ll_id\x18\x02 \x01(\r\x12!\n\x07\x65ntries\x18\x03 \x03(\x0b\x32\x10.PowerDeltaEntry\"P\n\x12Teros12Measurement\x12\x0f\n\x07vwc_raw\x18\x02 \x01(\x01\x12\x0f\n\x07vwc_adj\x18\x03 \x01(\x01\x12\x0c\n\x04temp\x18\x04 \x01(\x01\x12\n\n\x02\x65\x63\x18\x05 \x01(\r\"6\n\x12Teros21Measurement\x12\x12\n\nmatric_pot\x18\x01 \x01(\x01\x12\x0c\n\x04temp\x18\x02 \x01(\x01\"<\n\x13Phytos31Measurement\x12\x0f\n\x07voltage\x18\x01 \x01(\x01\x12\x14\n\x0cleaf_wetness\x18\x02 \x01(\x01\"L\n\x11\x42ME280Measurement\x12\x10\n\x08pressure\x18\x01 \x01(\r\x12\x13\n\x0btemperature\x18\x02 \x01(\x05\x12\x10\n\x08humidity\x18\x03 \x01(\r\"7\n\x12SEN0308Measurement\x12\x0f\n\x07voltage\x18\x01 \x01(\x01\x12\x10\n\x08humidity\x18\x02 \x01(\x01\"7\n\x12SEN0257Measurement\x12\x0f\n\x07voltage\x18\x01 \x01(\x01\x12\x10\n\x08pressure\x18\x02 \x01(\x01\"\"\n\x12YFS210CMeasurement\x12\x0c\n\x04\x66low\x18\x01 \x01(\x01\"(\n\x11PCAP02Measurement\x12\x13\n\x0b\x63\x61pacitance\x18\x01 \x01(\x01\"\xa2\x03\n\x0bMeasurement\x12\"\n\x04meta\x18\x01 \x01(\x0b\x32\x14.MeasurementMetadata\x12\"\n\x05power\x18\x02 \x01(\x0b\x32\x11.PowerMeasurementH\x00\x12&\n\x07teros12\x18\x03 \x01(\x0b\x32\x13.Teros12MeasurementH\x00\x12(\n\x08phytos31\x18\x04 \x01(\x0b\x32\x14.Phytos31MeasurementH\x00\x12$\n\x06\x62me280\x18\x05 \x01(\x0b\x32\x12.BME280MeasurementH\x00\x12&\n\x07teros21\x18\x06 \x01(\x0b\x32\x13.Teros21MeasurementH\x00\x12&\n\x07sen0308\x18\x07 \x01(\x0b\x32\x13.SEN0308MeasurementH\x00\x12&\n\x07sen0257\x18\x08 \x01(\x0b\x32\x13.SEN0257MeasurementH\x00\x12&\n\x07yfs210c\x18\t \x01(\x0b\x32\x13.YFS210CMeasurementH\x00\x12$\n\x06pcap02\x18\n \x01(\x0b\x32\x12.PCAP02MeasurementH\x00\x42\r\n\x0bmeasurement\"X\n\x08Response\x12$\n\x04resp\x18\x01 \x01(\x0e\x32\x16.Response.ResponseType\"&\n\x0cResponseType\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"\xc4\x02\n\x0c\x45sp32Command\x12$\n\x0cpage_command\x18\x01 \x01(\x0b\x32\x0c.PageCommandH\x00\x12$\n\x0ctest_command\x18\x02 \x01(\x0b\x32\x0c.TestCommandH\x00\x12$\n\x0cwifi_command\x18\x03 \x01(\x0b\x32\x0c.WiFiCommandH\x00\x12*\n\x0fmicrosd_command\x18\x04 \x01(\x0b\x32\x0f.MicroSDCommandH\x00\x12\x30\n\x12irrigation_command\x18\x05 \x01(\x0b\x32\x12.IrrigationCommandH\x00\x12\x31\n\x13user_config_command\x18\x06 \x01(\x0b\x32\x12.UserConfigCommandH\x00\x12&\n\rpower_command\x18\x07 \x01(\x0b\x32\r.PowerCommandH\x00\x42\t\n\x07\x63ommand\"\xb6\x01\n\x0bPageCommand\x12.\n\x0c\x66ile_request\x18\x01 \x01(\x0e\x32\x18.PageCommand.RequestType\x12\x17\n\x0f\x66ile_descriptor\x18\x02 \x01(\r\x12\x12\n\nblock_size\x18\x03 \x01(\r\x12\x11\n\tnum_bytes\x18\x04 \x01(\r\"7\n\x0bRequestType\x12\x08\n\x04OPEN\x10\x00\x12\t\n\x05\x43LOSE\x10\x01\x12\x08\n\x04READ\x10\x02\x12\t\n\x05WRITE\x10\x03\"\x82\x01\n\x0bTestCommand\x12\'\n\x05state\x18\x01 \x01(\x0e\x32\x18.TestCommand.ChangeState\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x05\"<\n\x0b\x43hangeState\x12\x0b\n\x07RECEIVE\x10\x00\x12\x13\n\x0fRECEIVE_REQUEST\x10\x01\x12\x0b\n\x07REQUEST\x10\x02\"\xc5\x02\n\x0bWiFiCommand\x12\x1f\n\x04type\x18\x01 \x01(\x0e\x32\x11.WiFiCommand.Type\x12\x0c\n\x04ssid\x18\x02 \x01(\t\x12\x0e\n\x06passwd\x18\x03 \x01(\t\x12\x0b\n\x03url\x18\x04 \x01(\t\x12\x0c\n\x04port\x18\x08 \x01(\r\x12\n\n\x02rc\x18\x05 \x01(\r\x12\n\n\x02ts\x18\x06 \x01(\r\x12\x0c\n\x04resp\x18\x07 \x01(\x0c\x12\x0b\n\x03mac\x18\t \x01(\t\x12\x0f\n\x07\x63lients\x18\n \x01(\r\"\x97\x01\n\x04Type\x12\x0b\n\x07\x43ONNECT\x10\x00\x12\x08\n\x04POST\x10\x01\x12\t\n\x05\x43HECK\x10\x02\x12\x08\n\x04TIME\x10\x03\x12\x0e\n\nDISCONNECT\x10\x04\x12\x0e\n\nCHECK_WIFI\x10\x05\x12\r\n\tCHECK_API\x10\x06\x12\x0c\n\x08NTP_SYNC\x10\x07\x12\x08\n\x04HOST\x10\x08\x12\r\n\tSTOP_HOST\x10\t\x12\r\n\tHOST_INFO\x10\n\"\xad\x01\n\x11UserConfigCommand\x12,\n\x04type\x18\x01 \x01(\x0e\x32\x1e.UserConfigCommand.RequestType\x12\'\n\x0b\x63onfig_data\x18\x02 \x01(\x0b\x32\x12.UserConfiguration\"A\n\x0bRequestType\x12\x12\n\x0eREQUEST_CONFIG\x10\x00\x12\x13\n\x0fRESPONSE_CONFIG\x10\x01\x12\t\n\x05START\x10\x02\"\x86\x03\n\x0eMicroSDCommand\x12\"\n\x04type\x18\x01 \x01(\x0e\x32\x14.MicroSDCommand.Type\x12\x10\n\x08\x66ilename\x18\x02 \x01(\t\x12&\n\x02rc\x18\x03 \x01(\x0e\x32\x1a.MicroSDCommand.ReturnCode\x12\x1c\n\x04meas\x18\x04 \x01(\x0b\x32\x0c.MeasurementH\x00\x12 \n\x02uc\x18\x05 \x01(\x0b\x32\x12.UserConfigurationH\x00\" \n\x04Type\x12\x08\n\x04SAVE\x10\x00\x12\x0e\n\nUSERCONFIG\x10\x01\"\xab\x01\n\nReturnCode\x12\x0b\n\x07SUCCESS\x10\x00\x12\x11\n\rERROR_GENERAL\x10\x01\x12\x1e\n\x1a\x45RROR_MICROSD_NOT_INSERTED\x10\x02\x12#\n\x1f\x45RROR_FILE_SYSTEM_NOT_MOUNTABLE\x10\x03\x12\x1d\n\x19\x45RROR_PAYLOAD_NOT_DECODED\x10\x04\x12\x19\n\x15\x45RROR_FILE_NOT_OPENED\x10\x05\x42\x06\n\x04\x64\x61ta\"\x94\x01\n\x11IrrigationCommand\x12%\n\x04type\x18\x01 \x01(\x0e\x32\x17.IrrigationCommand.Type\x12\'\n\x05state\x18\x02 \x01(\x0e\x32\x18.IrrigationCommand.State\"\x11\n\x04Type\x12\t\n\x05\x43HECK\x10\x00\"\x1c\n\x05State\x12\x08\n\x04OPEN\x10\x00\x12\t\n\x05\x43LOSE\x10\x01\"\xab\x03\n\x0cPowerCommand\x12 \n\x04type\x18\x01 \x01(\x0e\x32\x12.PowerCommand.Type\x12*\n\x06reason\x18\x02 \x01(\x0e\x32\x1a.PowerCommand.WakeupReason\x12\x12\n\nboot_count\x18\x03 \x01(\r\"\x1d\n\x04Type\x12\t\n\x05SLEEP\x10\x00\x12\n\n\x06WAKEUP\x10\x01\"\x99\x02\n\x0cWakeupReason\x12\x15\n\x11POWER_WAKEUP_EXT0\x10\x00\x12\x15\n\x11POWER_WAKEUP_EXT1\x10\x01\x12\x16\n\x12POWER_WAKEUP_TIMER\x10\x02\x12\x19\n\x15POWER_WAKEUP_TOUCHPAD\x10\x03\x12\x14\n\x10POWER_WAKEUP_ULP\x10\x04\x12\x15\n\x11POWER_WAKEUP_GPIO\x10\x05\x12\x15\n\x11POWER_WAKEUP_UART\x10\x06\x12\x15\n\x11POWER_WAKEUP_WIFI\x10\x07\x12\x16\n\x12POWER_WAKEUP_COCPU\x10\x08\x12 \n\x1cPOWER_WAKEUP_COCPU_TRAP_TRIG\x10\t\x12\x13\n\x0fPOWER_WAKEUP_BT\x10\n\"\xdc\x02\n\x11UserConfiguration\x12\x11\n\tlogger_id\x18\x01 \x01(\r\x12\x0f\n\x07\x63\x65ll_id\x18\x02 \x01(\r\x12$\n\rUpload_method\x18\x03 \x01(\x0e\x32\r.Uploadmethod\x12\x17\n\x0fUpload_interval\x18\x04 \x01(\r\x12\'\n\x0f\x65nabled_sensors\x18\x05 \x03(\x0e\x32\x0e.EnabledSensor\x12\x15\n\rVoltage_Slope\x18\x06 \x01(\x01\x12\x16\n\x0eVoltage_Offset\x18\x07 \x01(\x01\x12\x15\n\rCurrent_Slope\x18\x08 \x01(\x01\x12\x16\n\x0e\x43urrent_Offset\x18\t \x01(\x01\x12\x11\n\tWiFi_SSID\x18\n \x01(\t\x12\x15\n\rWiFi_Password\x18\x0b \x01(\t\x12\x18\n\x10\x41PI_Endpoint_URL\x18\x0c \x01(\t\x12\x19\n\x11\x41PI_Endpoint_Port\x18\r \x01(\r\"\x17\n\x08\x61\x64\x63Value\x12\x0b\n\x03\x61\x64\x63\x18\x01 \x01(\r*\x90\x01\n\rEnabledSensor\x12\x0b\n\x07Voltage\x10\x00\x12\x0b\n\x07\x43urrent\x10\x01\x12\x0b\n\x07Teros12\x10\x02\x12\x0b\n\x07Teros21\x10\x03\x12\n\n\x06\x42ME280\x10\x04\x12\x0c\n\x08Phytos31\x10\x05\x12\x0b\n\x07SEN0308\x10\x06\x12\x0b\n\x07SEN0257\x10\x07\x12\x0b\n\x07YFS210C\x10\x08\x12\n\n\x06PCAP02\x10\t*\"\n\x0cUploadmethod\x12\x08\n\x04LoRa\x10\x00\x12\x08\n\x04WiFi\x10\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'soil_power_sensor_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_ENABLEDSENSOR']._serialized_start=4041
  _globals['_ENABLEDSENSOR']._serialized_end=4185
  _globals['_UPLOADMETHOD']._serialized_start=4187
  _globals['_UPLOADMETHOD']._serialized_end=4221
  _globals['_MEASUREMENTMETADATA']._serialized_start=27
  _globals['_MEASUREMENTMETADATA']._serialized_end=96
  _globals['_POWERMEASUREMENT']._serialized_start=98
  _globals['_POWERMEASUREMENT']._serialized_end=150
  _globals['_VOLTAGEDELTAMEASUREMENT']._serialized_start=152
  _globals['_VOLTAGEDELTAMEASUREMENT']._serialized_end=194
  _globals['_CURRENTDELTAMEASUREMENT']._serialized_start=196
  _globals['_CURRENTDELTAMEASUREMENT']._serialized_end=238
  _globals['_VOLTAGEMEASUREMENT']._serialized_start=240
  _globals['_VOLTAGEMEASUREMENT']._serialized_end=277
  _globals['_CURRENTMEASUREMENT']._serialized_start=279
  _globals['_CURRENTMEASUREMENT']._serialized_end=316
  _globals['_POWERDELTAENTRY']._serialized_start=318
  _globals['_POWERDELTAENTRY']._serialized_end=393
  _globals['_POWERMEASUREMENTDELTA']._serialized_start=395
  _globals['_POWERMEASUREMENTDELTA']._serialized_end=464
  _globals['_REPEATEDPOWERDELTAS']._serialized_start=466
  _globals['_REPEATEDPOWERDELTAS']._serialized_end=558
  _globals['_TEROS12MEASUREMENT']._serialized_start=560
  _globals['_TEROS12MEASUREMENT']._serialized_end=640
  _globals['_TEROS21MEASUREMENT']._serialized_start=642
  _globals['_TEROS21MEASUREMENT']._serialized_end=696
  _globals['_PHYTOS31MEASUREMENT']._serialized_start=698
  _globals['_PHYTOS31MEASUREMENT']._serialized_end=758
  _globals['_BME280MEASUREMENT']._serialized_start=760
  _globals['_BME280MEASUREMENT']._serialized_end=836
  _globals['_SEN0308MEASUREMENT']._serialized_start=838
  _globals['_SEN0308MEASUREMENT']._serialized_end=893
  _globals['_SEN0257MEASUREMENT']._serialized_start=895
  _globals['_SEN0257MEASUREMENT']._serialized_end=950
  _globals['_YFS210CMEASUREMENT']._serialized_start=952
  _globals['_YFS210CMEASUREMENT']._serialized_end=986
  _globals['_PCAP02MEASUREMENT']._serialized_start=988
  _globals['_PCAP02MEASUREMENT']._serialized_end=1028
  _globals['_MEASUREMENT']._serialized_start=1031
  _globals['_MEASUREMENT']._serialized_end=1449
  _globals['_RESPONSE']._serialized_start=1451
  _globals['_RESPONSE']._serialized_end=1539
  _globals['_RESPONSE_RESPONSETYPE']._serialized_start=1501
  _globals['_RESPONSE_RESPONSETYPE']._serialized_end=1539
  _globals['_ESP32COMMAND']._serialized_start=1542
  _globals['_ESP32COMMAND']._serialized_end=1866
  _globals['_PAGECOMMAND']._serialized_start=1869
  _globals['_PAGECOMMAND']._serialized_end=2051
  _globals['_PAGECOMMAND_REQUESTTYPE']._serialized_start=1996
  _globals['_PAGECOMMAND_REQUESTTYPE']._serialized_end=2051
  _globals['_TESTCOMMAND']._serialized_start=2054
  _globals['_TESTCOMMAND']._serialized_end=2184
  _globals['_TESTCOMMAND_CHANGESTATE']._serialized_start=2124
  _globals['_TESTCOMMAND_CHANGESTATE']._serialized_end=2184
  _globals['_WIFICOMMAND']._serialized_start=2187
  _globals['_WIFICOMMAND']._serialized_end=2512
  _globals['_WIFICOMMAND_TYPE']._serialized_start=2361
  _globals['_WIFICOMMAND_TYPE']._serialized_end=2512
  _globals['_USERCONFIGCOMMAND']._serialized_start=2515
  _globals['_USERCONFIGCOMMAND']._serialized_end=2688
  _globals['_USERCONFIGCOMMAND_REQUESTTYPE']._serialized_start=2623
  _globals['_USERCONFIGCOMMAND_REQUESTTYPE']._serialized_end=2688
  _globals['_MICROSDCOMMAND']._serialized_start=2691
  _globals['_MICROSDCOMMAND']._serialized_end=3081
  _globals['_MICROSDCOMMAND_TYPE']._serialized_start=2867
  _globals['_MICROSDCOMMAND_TYPE']._serialized_end=2899
  _globals['_MICROSDCOMMAND_RETURNCODE']._serialized_start=2902
  _globals['_MICROSDCOMMAND_RETURNCODE']._serialized_end=3073
  _globals['_IRRIGATIONCOMMAND']._serialized_start=3084
  _globals['_IRRIGATIONCOMMAND']._serialized_end=3232
  _globals['_IRRIGATIONCOMMAND_TYPE']._serialized_start=3185
  _globals['_IRRIGATIONCOMMAND_TYPE']._serialized_end=3202
  _globals['_IRRIGATIONCOMMAND_STATE']._serialized_start=3204
  _globals['_IRRIGATIONCOMMAND_STATE']._serialized_end=3232
  _globals['_POWERCOMMAND']._serialized_start=3235
  _globals['_POWERCOMMAND']._serialized_end=3662
  _globals['_POWERCOMMAND_TYPE']._serialized_start=3349
  _globals['_POWERCOMMAND_TYPE']._serialized_end=3378
  _globals['_POWERCOMMAND_WAKEUPREASON']._serialized_start=3381
  _globals['_POWERCOMMAND_WAKEUPREASON']._serialized_end=3662
  _globals['_USERCONFIGURATION']._serialized_start=3665
  _globals['_USERCONFIGURATION']._serialized_end=4013
  _globals['_ADCVALUE']._serialized_start=4015
  _globals['_ADCVALUE']._serialized_end=4038
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: sensor.proto
# Protobuf Python Version: 6.33.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    1,
    '',
    'sensor.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0csensor.proto\":\n\x08Metadata\x12\x0f\n\x07\x63\x65ll_id\x18\x01 \x01(\r\x12\x11\n\tlogger_id\x18\x02 \x01(\r\x12\n\n\x02ts\x18\x03 \x01(\r\":\n\x0eSensorResponse\x12\x0b\n\x03idx\x18\x01 \x01(\r\x12\x1b\n\x05\x65rror\x18\x02 \x01(\x0e\x32\x0c.SensorError\"=\n\x17RepeatedSensorResponses\x12\"\n\tresponses\x18\x01 \x03(\x0b\x32\x0f.SensorResponse\"\x9e\x01\n\x11SensorMeasurement\x12\x17\n\x04meta\x18\x01 \x01(\x0b\x32\t.Metadata\x12\x19\n\x04type\x18\x02 \x01(\x0e\x32\x0b.SensorType\x12\x16\n\x0cunsigned_int\x18\x03 \x01(\rH\x00\x12\x14\n\nsigned_int\x18\x04 \x01(\x05H\x00\x12\x11\n\x07\x64\x65\x63imal\x18\x05 \x01(\x01H\x00\x12\x0b\n\x03idx\x18\x06 \x01(\rB\x07\n\x05value\"z\n\x1aRepeatedSensorMeasurements\x12\x17\n\x04meta\x18\x01 \x01(\x0b\x32\t.Metadata\x12\x19\n\x04type\x18\x02 \x01(\x0e\x32\x0b.SensorType\x12(\n\x0cmeasurements\x18\x03 \x03(\x0b\x32\x12.SensorMeasurement*\x9a\x03\n\nSensorType\x12\x08\n\x04NONE\x10\x00\x12\x11\n\rPOWER_VOLTAGE\x10\x01\x12\x11\n\rPOWER_CURRENT\x10\x02\x12\x0f\n\x0bTEROS12_VWC\x10\x03\x12\x13\n\x0fTEROS12_VWC_ADJ\x10\x04\x12\x10\n\x0cTEROS12_TEMP\x10\x05\x12\x0e\n\nTEROS12_EC\x10\x06\x12\x14\n\x10PHYTOS31_VOLTAGE\x10\x07\x12\x19\n\x15PHYTOS31_LEAF_WETNESS\x10\x08\x12\x13\n\x0f\x42ME280_PRESSURE\x10\t\x12\x0f\n\x0b\x42ME280_TEMP\x10\n\x12\x13\n\x0f\x42ME280_HUMIDITY\x10\x0b\x12\x16\n\x12TEROS21_MATRIC_POT\x10\x0c\x12\x10\n\x0cTEROS21_TEMP\x10\r\x12\x13\n\x0fSEN0308_VOLTAGE\x10\x0e\x12\x14\n\x10SEN0308_HUMIDITY\x10\x0f\x12\x13\n\x0fSEN0257_VOLTAGE\x10\x10\x12\x14\n\x10SEN0257_PRESSURE\x10\x11\x12\x10\n\x0cYFS210C_FLOW\x10\x12\x12\x16\n\x12PCAP02_CAPACITANCE\x10\x13*b\n\x0bSensorError\x12\x06\n\x02OK\x10\x00\x12\x0b\n\x07GENERAL\x10\x01\x12\n\n\x06LOGGER\x10\x02\x12\x08\n\x04\x43\x45LL\x10\x03\x12\x0f\n\x0bUNSUPPORTED\x10\x04\x12\x0b\n\x07INVALID\x10\x05\x12\n\n\x06\x44\x45\x43ODE\x10\x06\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sensor_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SENSORTYPE']._serialized_start=485
  _globals['_SENSORTYPE']._serialized_end=895
  _globals['_SENSORERROR']._serialized_start=897
  _globals['_SENSORERROR']._serialized_end=995
  _globals['_METADATA']._serialized_start=16
  _globals['_METADATA']._serialized_end=74
  _globals['_SENSORRESPONSE']._serialized_start=76
  _globals['_SENSORRESPONSE']._serialized_end=134
  _globals['_REPEATEDSENSORRESPONSES']._serialized_start=136
  _globals['_REPEATEDSENSORRESPONSES']._serialized_end=197
  _globals['_SENSORMEASUREMENT']._serialized_start=200
  _globals['_SENSORMEASUREMENT']._serialized_end=358
  _globals['_REPEATEDSENSORMEASUREMENTS']._serialized_start=360
  _globals['_REPEATEDSENSORMEASUREMENTS']._serialized_end=482
# @@protoc_insertion_point(module_scope)



"""Module for sensor measurements

Encode/decoding functions are wrappers around the protobuf messages. These take
in the json dictionary format of the messages and return serialized byte
arrays.

Format/parse functions implement the protocol for repeated sensor measurements.
These take in a list of measurements and automatically optimize repeated
metadata fields.
"""

from google.protobuf.json_format import MessageToDict, ParseDict

from .sensor_pb2 import (
    SensorMeasurement,
    RepeatedSensorMeasurements,
    SensorType,
    RepeatedSensorResponses,
)


def parse_sensor_measurement(data: bytes) -> list:
    """Parses a sensor measurement into a usable dictionary.

    Function does the following:
        1. Decodes the serialized byte array
        2. Updates metadata for each measurement if missing
        3. Adds names, descriptions, and units to metadata

    Args:
        data: Byte array of message.

    Returns:
        Dictionary of sensor measurement.
    """

    meas = decode_repeated_sensor_measurements(data)
    meas = update_repeated_metadata(meas)
    for m in meas["measurements"]:
        sensor_data = get_sensor_data(m["type"])
        m.update(sensor_data)

    return meas


def format_sensor_measurement(meas: list) -> bytes:
    """Formats a sensor measurement dictionary into a serialized byte array.

    Function does the following:
        1. Uses top level metadata for duplicate measurement metadata fields
        2. Encodes the dictionary into a serialized byte array

    Args:
        meas: Dictionary of sensor measurement.

    Returns:
        Byte array of serialized message.
    """

    # TODO Implement optimization of repeated metadata fields
    meas_dict = {
        "measurements": meas,
    }

    data = encode_repeated_sensor_measurements(meas_dict)
    return data


def get_sensor_data(meas_type: int) -> dict:
    """Gets sensor data information.

    Args:
        meas_type: Sensor measurement dictionary.

    Returns:
        Metadata associated with the sensor type.
    """

    SENSOR_DATA = {
        SensorType.POWER_VOLTAGE: {
            "name": "Voltage",
            "unit": "mV",
        },
        SensorType.POWER_CURRENT: {
            "name": "Current",
            "unit": "uA",
        },
        SensorType.TEROS12_VWC: {
            "name": "Volumetric Water Content",
            "unit": "%",
        },
        SensorType.TEROS12_TEMP: {
            "name": "Temperature",
            "unit": "C",
        },
        SensorType.TEROS12_EC: {
            "name": "Electrical Conductivity",
            "unit": "uS/cm",
        },
        SensorType.PHYTOS31_VOLTAGE: {
            "name": "Voltage",
            "unit": "mV",
        },
        SensorType.PHYTOS31_LEAF_WETNESS: {
            "name": "Leaf Wetness",
            "unit": "%",
        },
        SensorType.BME280_PRESSURE: {
            "name": "Pressure",
            "unit": "kPa",
        },
        SensorType.BME280_TEMP: {
            "name": "Temperature",
            "unit": "C",
        },
        SensorType.BME280_HUMIDITY: {
            "name": "Humidity",
            "unit": "%",
        },
        SensorType.TEROS21_MATRIC_POT: {
            "name": "Matric Potential",
            "unit": "kPa",
        },
        SensorType.TEROS21_TEMP: {
            "name": "Temperature",
            "unit": "C",
        },
        SensorType.SEN0308_VOLTAGE: {
            "name": "Voltage",
            "unit": "mV",
        },
        SensorType.SEN0308_HUMIDITY: {
            "name": "Humidity",
            "unit": "%",
        },
        SensorType.SEN0257_VOLTAGE: {
            "name": "Voltage",
            "unit": "mV",
        },
        SensorType.SEN0257_PRESSURE: {
            "name": "Pressure",
            "unit": "kPa",
        },
        SensorType.YFS210C_FLOW: {
            "name": "Flow Rate",
            "unit": "L/min",
        },
    }

    meta = SENSOR_DATA[SensorType.Value(meas_type)]
    return meta


def encode_sensor_measurement(meas_dict: dict) -> bytes:
    meas = SensorMeasurement()
    ParseDict(meas_dict, meas)

    return meas.SerializeToString()


def encode_repeated_sensor_measurements(meas_dict: dict) -> bytes:
    """Encodes a SensorMeasurement message

    Args:
        rep_meas: Repeated sensor measurement dictionary.

    Returns:
        Byte array of encoded RepeatedSensorMeasurements message.
    """

    meas = RepeatedSensorMeasurements()
    ParseDict(meas_dict, meas)

    return meas.SerializeToString()


def decode_sensor_measurement(data: bytes) -> dict:
    """Decodes a SensorMeasurement message

    Args:
        data: Byte array of SensorMeasurement message.

    Returns:
        Decoded sensor measurement dictionary.
    """

    meas = SensorMeasurement()
    meas.ParseFromString(data)

    parsed_meas = MessageToDict(meas)

    return parsed_meas


def decode_repeated_sensor_measurements(data: bytes) -> dict:
    """Decodes repeated sensor measurements

    Args:
        data: Byte array from RepeatedSensorMeasurements

    Returns:
        List of decoded sensor measurement dictionaries.
    """

    rep_meas = RepeatedSensorMeasurements()
    rep_meas.ParseFromString(data)
    return MessageToDict(rep_meas)


def update_repeated_metadata(meas: dict) -> dict:
    """Ensures every measurements has metadata field set.

    If a measurement is missing the metadata field, it is filled in from the
    repeated sensor measurement. Existing measurement metadata fields are not
    overwritten.

    Args:
        meas: Sensor measurement dictionary.

    Returns:
        Updated sensor measurement dictionary.
    """

    # if top level meta does not exist, ensure all measurements have meta
    if "meta" not in meas:
        for m in meas["measurements"]:
            if "meta" not in m:
                raise ValueError("Repeated measurement missing metadata field.")
    # otherwise populate missing measurement meta from top level
    else:
        for m in meas["measurements"]:
            if "meta" not in m:
                m["meta"] = meas["meta"]
        del meas["meta"]

    return meas


def encode_sensor_response(resp_dict: dict) -> bytes:
    """Encodes a sensor response message.

    {
        responses: [
            {
                status: int,
                message: str,
            },
            ...
        ]
    }

    Args:
        resp_dict: Sensor response dictionary.

    Returns:
        Byte array of encoded SensorResponse message.
    """

    resp = RepeatedSensorResponses()
    ParseDict(resp_dict, resp)

    return resp.SerializeToString()


def decode_sensor_response(data: bytes) -> dict:
    """Decodes a sensor response message.

    {
        responses: [
            {
                status: int,
                message: str,
            },
            ...
        ]
    }

    Args:
        data: Byte array of SensorResponse message.

    Returns:
        Decoded sensor response dictionary.
    """

    resp = RepeatedSensorResponses()
    resp.ParseFromString(data)

    parsed_resp = MessageToDict(resp)

    return parsed_resp



"""Module to decode soil power sensor messages"""

from google.protobuf.json_format import MessageToDict

from .soil_power_sensor_pb2 import Measurement, Response, UserConfiguration


def decode_response(data: bytes):
    """Decodes a Response message.

    Args:
        data: Byte array of Response message.

    Returns:
        Returns the ResponseType.

    Raises:
        KeyError: Missing the resp field.
    """

    resp = Response()
    resp.ParseFromString(data)

    if not resp.HasField("resp"):
        raise KeyError("Missing response type")

    return resp.resp


def decode_measurement(data: bytes, raw: bool = True) -> dict:
    """Decodes a Measurement message

    The data is decoded into a flat dictionary with the measurement type.

    Args:
        data: Byte array of Measurement message.
        raw: Flag to return raw or adjusted measurements

    Returns:
        Flat dictionary of values from the meta field, measurement field, and
        the key "type" to indicate the type of measurement.

    Raises:
        KeyError: When the serialized data is missing a required field.
    """

    # parse data
    meas = Measurement()
    meas.ParseFromString(data)

    # convert meta into dict
    if not meas.HasField("meta"):
        raise KeyError("Measurement missing metadata")
    meta_dict = MessageToDict(meas.meta, always_print_fields_with_no_presence=True)

    # decode measurement
    if not meas.HasField("measurement"):
        raise KeyError("Measurement missing data")
    measurement_type = meas.WhichOneof("measurement")
    measurement_dict = MessageToDict(
        getattr(meas, measurement_type), always_print_fields_with_no_presence=True
    )

    # store measurement type
    meta_dict["type"] = measurement_type

    # store measurement data
    meta_dict["data"] = measurement_dict

    # process raw
    if not raw:
        # convert measurements to hPa, C, and %
        if meta_dict["type"] == "bme280":
            meta_dict["data"]["pressure"] /= 10.0
            meta_dict["data"]["temperature"] /= 100.0
            meta_dict["data"]["humidity"] /= 1000.0

    # store measurement type
    meta_dict["data_type"] = {}
    for key, value in measurement_dict.items():
        meta_dict["data_type"][key] = type(value)
    return meta_dict


def decode_user_configuration(data: bytes) -> dict:
    """Decodes a UserConfiguration message

    Args:
        data: Byte array of UserConfiguration message.

    Returns:
        Dictionary of UserConfiguration values.

    Raises:
        KeyError: When the serialized data is missing a required field.
    """

    user_config = UserConfiguration()
    user_config.ParseFromString(data)

    if user_config.cell_id == 0 or user_config.logger_id == 0:
        raise KeyError("User configuration missing required fields")

    user_config_dict = MessageToDict(
        user_config, always_print_fields_with_no_presence=True
    )

    return user_config_dict



from google.protobuf import message
from google.protobuf.json_format import MessageToDict

from .soil_power_sensor_pb2 import Esp32Command, PageCommand, TestCommand, WiFiCommand


def encode_esp32command(cmd_type: str, **kwargs) -> bytes:
    """Encodes a command for the esp32

    Args:
        cmd: Type of command to encode
        **kwargs: See other encode functions

    Returns:
        Bytes of protobuf serialized message
    """

    cmd = Esp32Command()

    if cmd_type == "page":
        page_cmd = encode_page_command(**kwargs)
        cmd.page_command.CopyFrom(page_cmd)
    elif cmd_type == "test":
        test_cmd = encode_test_command(**kwargs)
        cmd.test_command.CopyFrom(test_cmd)
    elif cmd_type == "wifi":
        wifi_cmd = encode_wifi_command(**kwargs)
        cmd.wifi_command.CopyFrom(wifi_cmd)
    else:
        raise NotImplementedError(f"Command type {cmd_type} not implemented!")

    return cmd.SerializeToString()


def encode_page_command(req: str, fd: int, bs: int = 0, n: int = 0) -> message:
    """Encodes a command for memory paging

    Args:
        req: Page request type
        fd: File descriptor
        bs: Block size
        n: Number of bytes

    Return:
        PageCommand message

    Raises:
        NotImplementedError: When page request type does not exist
    """

    # look up table to request types
    req_lut = {
        "open": PageCommand.RequestType.OPEN,
        "close": PageCommand.RequestType.CLOSE,
        "read": PageCommand.RequestType.READ,
        "write": PageCommand.RequestType.WRITE,
    }

    # ensure lower case
    req = req.lower()

    page_cmd = PageCommand()
    try:
        page_cmd.file_request = req_lut[req]
    except KeyError as exc:
        raise NotImplementedError(f"File request type {req} not implemented") from exc
    page_cmd.file_descriptor = fd
    page_cmd.block_size = bs
    page_cmd.num_bytes = n

    return page_cmd


def encode_test_command(state: str, data: int) -> message:
    """Encodes a command for testing modules library

    The following is of valid strings for state:
        "receive": Data is received by the module
        "receive_request": Data is received by the module indicating data for
                           subsequent request
        "request": Data is sent my the module

    Args:
        state (str): State command string
        data (int): Integer to test passing of data

    Return:
        TestCommand message

    Raises:
        NotImplementedError: When the ChangeState type does not exist
    """

    # look up table for state strings
    cs_lut = {
        "receive": TestCommand.ChangeState.RECEIVE,
        "receive_request": TestCommand.ChangeState.RECEIVE_REQUEST,
        "request": TestCommand.ChangeState.REQUEST,
    }

    # ensure all lower case
    state = state.lower()

    test_cmd = TestCommand()
    try:
        test_cmd.state = cs_lut[state]
    except KeyError as exc:
        raise NotImplementedError(f"State command {state} not implemented") from exc
    test_cmd.data = data

    return test_cmd


def encode_wifi_command(
    _type: str,
    ssid: str = "",
    passwd: str = "",
    url: str = "",
    port: int = 0,
    rc: int = 0,
    ts: int = 0,
    resp: bytes = b"",
) -> message:
    """Encodes a WiFi command

    Valid strings for _type are:
        "connect": Connecting to a WiFi network and any other setup
        "post": Posting data to the endpoint

    Not all arguments are necessary for valid commands.

    Args:
        _type; Type of request
        ssid: WiFi SSID
        passwd: WiFi password
        url: Endpoint URL
        port: Port of webserver
        rc: Return code
        ts: Timestamp
        resp: Binary data response from server

    Returns:
        WiFiCommand message

    Raises:
        NotImplementedError: When _type does not exist
    """

    # look up table for _type strings
    _type_lut = {"connect": WiFiCommand.Type.CONNECT, "post": WiFiCommand.Type.POST}

    # ensure all lower case
    _type = _type.lower()

    wifi_cmd = WiFiCommand()
    try:
        wifi_cmd.type = _type_lut[_type]
    except KeyError as exc:
        raise NotImplementedError(f"Type {_type} is not implemented") from exc

    wifi_cmd.ssid = ssid
    wifi_cmd.passwd = passwd
    wifi_cmd.url = url
    wifi_cmd.port = port
    wifi_cmd.rc = rc
    wifi_cmd.ts = ts
    wifi_cmd.resp = resp

    return wifi_cmd


def decode_esp32command(data: bytes) -> dict:
    cmd = Esp32Command()

    # parse serialized message
    cmd.ParseFromString(data)

    # return dict
    return MessageToDict(cmd, always_print_fields_with_no_presence=True)



"""Testing simulation code

Tests the functionality of the NodeSimulator class uploading data to a sample
server. A HTTP server to handle post request is started on a per test basis at
localhost:8080. Ensure there is not another process using this port.
"""

import unittest
from http.server import BaseHTTPRequestHandler, HTTPServer

from multiprocessing import Process
from datetime import datetime, timedelta
from time import sleep

from ents.simulator.node import NodeSimulator


class Backend(BaseHTTPRequestHandler):
    """Backend simulator"""

    def log_message(self, format, *args):
        """Override to not print to stdout"""
        pass

    def do_POST(self):
        """Return 200 from a POST request"""

        # print request
        # content_length = int(self.headers["Content-Length"])
        # data = self.rfile.read(content_length)

        # send response
        self.send_response(200)
        self.send_header("Content-type", "text/octet-stream")
        self.end_headers()
        self.wfile.write(b"The world is your oyster!")


def run_server():
    """Run the simulated HTTP server on localhost:8080"""

    # run server
    server = HTTPServer(("localhost", 8080), Backend)
    server.serve_forever()


class TestNodeSimulator(unittest.TestCase):
    url = "http://localhost:8080/"

    def setUp(self):
        """Start webserver in separate process"""

        self.backend = Process(target=run_server)
        self.backend.start()
        # give server time to start
        sleep(0.5)

    def tearDown(self):
        """Terminate process"""

        self.backend.terminate()

    def test_power(self):
        sim = NodeSimulator(1, 2, "power")
        ts = int(datetime(2025, 5, 7).timestamp())
        sim.measure(ts)
        sim.send_next(self.url)

    def test_teros12(self):
        sim = NodeSimulator(1, 2, "teros12")
        ts = int(datetime(2025, 5, 7).timestamp())
        sim.measure(ts)
        sim.send_next(self.url)

    def test_teros21(self):
        sim = NodeSimulator(1, 2, "teros21")
        ts = int(datetime(2025, 5, 7).timestamp())
        sim.measure(ts)
        sim.send_next(self.url)

    def test_bme280(self):
        sim = NodeSimulator(1, 2, "bme280")
        ts = int(datetime(2025, 5, 7).timestamp())
        sim.measure(ts)
        sim.send_next(self.url)

    def test_max_connections(self):
        """Covers the case of streaming or batch mode with more than ~1024
        uploads. Can cause an error with the max number of file descriptors
        """

        sim = NodeSimulator(1, 2, "bme280")
        ts_start = datetime(2025, 5, 7)

        for i in range(2000):
            ts = int((ts_start + timedelta(days=i)).timestamp())
            sim.measure(ts)
            sim.send_next(self.url)


if __name__ == "__main__":
    # run unittests
    unittest.main()



"""
@brief testing advance trace receive in stm32
@file adv_trace.py
@author Ahmed Hassan Falah
@date 2024-10-12
"""

import serial
import serial.tools.list_ports


def sendToUART():
    """
    Sends the encoded configuration data via UART.
    """
    ser = None
    try:
        # Open the serial port
        ser = serial.Serial(port="COM6", baudrate=115200, timeout=2)
        # Send data (value 1)
        data = bytes([1])
        print(f"Sending: {data}")
        ser.write(data)
        print(
            "________________________________________________________________________"
        )
        print(f"{data}")

        # Read acknowledgment (1 byte)
        ack = ser.read(1)
        print(f"Received from STM32: {ack}")
        print(
            "________________________________________________________________________"
        )

        # Check acknowledgment
        if ack == b"A":
            print("Success")
        elif ack == b"N":
            print("Error NACK")
            return False
        else:
            print("Error")
            return False

    except serial.SerialException as e:
        print(f"UART Error: {e}")
        return False

    finally:
        if ser is not None:
            ser.close()


if __name__ == "__main__":
    sendToUART()
